"""Test connector."""

from typing import Tuple

import pytest

from tests import CONFIG_FILE_NM

from snowmobile import Snowmobile

CACHE_TESTING_ITEM_NAME1 = "test-path"
CACHE_TESTING_ITEM_NAME2 = "test-path2"


@pytest.fixture(scope="module")
def cache_with_a_testing_value_saved(sn_delayed) -> Tuple[Snowmobile, str]:
    """Performs setup for cache testing.

    Specifically:
        *   Uses current configuration file location as the (str) value to cache
        *   Ensures the name of the item to cache isn't already cached
        *   Saves a test item to the cache
        *   Returns the modified connector object and the value that was cached

    """
    # use current config file location as the value to cache
    value_to_cache = sn_delayed.cfg.location.as_posix()

    # ensure 'test-path' is not currently a cached value
    if sn_delayed.cfg.cache.get(CACHE_TESTING_ITEM_NAME1):
        sn_delayed.cfg.cache.clear(CACHE_TESTING_ITEM_NAME1)

    # save two values to cache under the testing item names
    for name in [CACHE_TESTING_ITEM_NAME1, CACHE_TESTING_ITEM_NAME2]:
        sn_delayed.cfg.cache.save_item(item_name=name, item_value=value_to_cache)

    return sn_delayed, value_to_cache


@pytest.mark.cache
@pytest.mark.configuration
def test_cache_save(cache_with_a_testing_value_saved):
    """Tests that the value cached matches the string value retrieved based on its name."""
    sn_delayed, cached_value = cache_with_a_testing_value_saved

    test_value_from_cache = sn_delayed.cfg.cache.get(CACHE_TESTING_ITEM_NAME1)

    assert cached_value == test_value_from_cache


@pytest.mark.cache
@pytest.mark.configuration
def test_cache_as_path(cache_with_a_testing_value_saved):
    """Tests the value cached matches the pathlib.Path object retrieved from cache.as_path()."""
    sn_delayed, _ = cache_with_a_testing_value_saved

    test_value_from_cache = sn_delayed.cfg.cache.as_path(CACHE_TESTING_ITEM_NAME1)

    assert sn_delayed.cfg.location == test_value_from_cache


@pytest.mark.cache
@pytest.mark.configuration
def test_cache_saving_and_deletion_operations(cache_with_a_testing_value_saved):
    """Test basic behavior of ``snowmobile.toml`` file export."""

    # connector containing a cache object, a sample path to cache
    sn_delayed, cached_path = cache_with_a_testing_value_saved

    # two distinct names to use when caching the sample path
    cached_names = [CACHE_TESTING_ITEM_NAME1, CACHE_TESTING_ITEM_NAME2]

    # ---

    # clearing cache of these items to set starting state for tests
    for item in cached_names:
        sn_delayed.cfg.cache.clear(item)
    # - asserting only to verify state is at it should be before rest of test -
    assert not sn_delayed.cfg.cache.contains(cached_names)

    # ---

    # save two items
    sn_delayed.cfg.cache.save_all({k: cached_path for k in cached_names})

    # verify that they **are** a part of the cache
    assert sn_delayed.cfg.cache.contains(cached_names)
    # re-verify but passing them to .contains() one item at a time
    for item in cached_names:
        assert sn_delayed.cfg.cache.contains(item)

    # ---

    # clear cache of these items again; ensures the assertion on starting state
    # was a result of the .clear() removing their contents, not their never being
    # cached in the first place
    sn_delayed.cfg.cache.clear(cached_names)
    # verify they are **not** part of the cache
    assert not sn_delayed.cfg.cache.contains(cached_names)

    # ---


@pytest.mark.cache
@pytest.mark.configuration
def test_cache_dunder_methods(cache_with_a_testing_value_saved):
    """Ensuring dunder str/repr methods throw no errors"""
    assert str(cache_with_a_testing_value_saved)
    assert repr(cache_with_a_testing_value_saved)


@pytest.mark.configuration
def test_export_configuration(tmpdir):
    """Test basic behavior of ``snowmobile.toml`` file export."""
    from pathlib import Path
    from snowmobile import Configuration

    tmpdir = Path(tmpdir)
    Configuration(export_dir=tmpdir)
    assert (tmpdir / "snowmobile.toml").exists()


@pytest.mark.io
@pytest.mark.configuration
def test_error_on_no_configuration_file_found():
    """Test basic behavior of ``snowmobile.toml`` file export."""
    from snowmobile import Configuration

    with pytest.raises(Exception):
        _ = Configuration(config_file_nm="_an_invalid_snowmobile_test_file.txt")


@pytest.mark.configuration
def test_finds_a_valid_configuration_file():
    """"""
    from snowmobile import Configuration

    cfg = Configuration(config_file_nm=CONFIG_FILE_NM)
    cfg.cache.clear(CONFIG_FILE_NM)

    cfg2 = Configuration(config_file_nm=CONFIG_FILE_NM)
    assert cfg.location == cfg2.location


@pytest.mark.configuration
def test_configuration_json_serialization():
    """Testing serialization methods of configuration objects."""
    from snowmobile import Configuration

    cfg = Configuration(config_file_nm=CONFIG_FILE_NM)

    assert cfg.json(by_alias=True) == cfg.__json__(by_alias=True)
    assert cfg.json(by_alias=False) != cfg.json(by_alias=True)


@pytest.mark.configuration
def test_base_json_serialization():
    """Testing serialization methods of pydantic-derived objects."""
    from snowmobile import Configuration

    cfg = Configuration(config_file_nm=CONFIG_FILE_NM)

    assert cfg.script.json(by_alias=True) == cfg.script.__json__(by_alias=True)
    assert cfg.script.json(by_alias=False) != cfg.script.json(by_alias=True)


@pytest.mark.configuration
def test_get_attrs_from_obj():
    """Tests 'attrs_from_obj() method of configuration class."""

    # noinspection PyMissingOrEmptyDocstring
    class AnyClass:
        """Example class with an attribute, a method, and a property."""

        def __init__(self):
            self.any_attr = 1

        def any_callable_method(self):
            return self.any_attr + 1

        @property
        def any_property_as_well(self):
            return self.any_callable_method() + 1

    from snowmobile import Configuration

    cfg = Configuration(config_file_nm=CONFIG_FILE_NM)

    any_instance_of_any_class = AnyClass()
    attributes = cfg.attrs_from_obj(obj=any_instance_of_any_class)
    methods = cfg.methods_from_obj(obj=any_instance_of_any_class)

    assert attributes["any_attr"] == 1
    assert methods["any_callable_method"]() == 2
    assert attributes["any_property_as_well"] == 3


@pytest.mark.configuration
def test_configuration_dunder_methods():
    """Verifies appearance of __str__ and __repr__ methods."""
    from snowmobile import Configuration

    cfg = Configuration(config_file_nm=CONFIG_FILE_NM)

    assert str(cfg) == f"snowmobile.Configuration('{CONFIG_FILE_NM}')"
    assert (
        str(repr(cfg)) == f"snowmobile.Configuration(config_file_nm='{CONFIG_FILE_NM}')"
    )


@pytest.mark.configuration
def test_set_item_on_base_class():
    """Verifies behavior of __setitem__ dunder method for base class."""
    from snowmobile import Configuration

    cfg = Configuration(config_file_nm=CONFIG_FILE_NM)
    cfg.script.any_arbitrary_attribute = 1  # cfg.script derives from Base, cfg does not
    cfg.script["tw"] = 1
    assert cfg.script.any_arbitrary_attribute
    assert cfg.script.tw



"""
:class:`SQL` contains utility methods to generate & execute
common SQL commands.

.. note::
   The :attr:`~SQL.auto_run` attribute defaults to `True`, meaning that the
   generated sql will execute when a method is called; if set to `False`
   the method will return the sql as a string without executing.

   The :class:`SQL` object is primarily interacted with as a
   pre-instantiated attribute of :class:`~snowmobile.Snowmobile`; in these
   instances users can fetch the generated sql as a string either by:

   1. Providing *run=False* to any method called; this will override
      all behavior set by the current value of :attr:`auto_run`
   2. Setting the :attr:`auto_run` attribute to `False` on an existing
      instance of :class:`SQL`, which will replicate the behavior of
      `(1)` without needing to provide *run=False* to each method
      called on that instance

"""
from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Optional, Union

import pandas as pd

from . import Generic
from .utils.parsing import p, strip, up


class SQL(Generic):
    """SQL class for generation & execution of common sql commands.

    Intended to be interacted with as an attribute of
    :class:`~snowmobile.Snowmobile`.

    note:
        *   All arguments except for :attr:`sn` are optional.
        *   The benefit of setting the other attributes on an instance of :class:`SQL`
            is to (optionally) avoid passing the same information to multiple methods
            when generating a variety of statements around the same object.

    Attributes:
        sn (snowmobile.Snowmobile):
            :class:`Snowmobile` for sql execution and connection information.
        nm (str):
            Object name to use in generated sql (e.g. 'some_table_name')
        obj (str):
            Object type to use in generated sql (e.g. 'table')
        schema (str):
            Schema to use when dot-prefixing sql; defaults to the schema with which the
            :attr:`sn` is connected to.
        auto_run (bool):
            Indicates whether to automatically execute the sql generated by a given
            method; defaults to *True*

    """

    def __init__(
        self,
        sn=None,
        nm: Optional[str] = None,
        obj: Optional[str] = None,
        auto_run: Optional[bool] = True,
    ):
        """Initializes a :class:`snowmobile.SQL` object."""
        super().__init__()
        self.sn = sn
        schema, nm = p(nm=nm)
        self.nm: str = nm
        self.schema = schema or sn.cfg.connection.current.schema_name
        self.obj: str = obj or "table"
        self.auto_run: bool = auto_run

    def info_schema_tables(
        self,
        nm: Optional[str] = None,
        fields: List[str] = None,
        restrictions: Dict[str, str] = None,
        order_by: List[Optional[str, int]] = None,
        all_schemas: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Query ``information_schema.tables`` for a given table or view.

        Args:
            nm (str):
                Table name, including schema if creating a stage outside of the
                current schema.
            fields (List[str]):
                List of fields to include in returned results (e.g.
                ['table_name', 'table_type', 'last_altered'])
            restrictions (List[str]):
                List of conditionals typed as literal components of a `where`
                clause (e.g.
                ["table_type = 'base table'", 'last_altered::date = current_date()']
                ).
            order_by (List[str]):
                List of fields or their ordinal positions to order the results by.
            all_schemas (bool):
                Include tables/views from all schemas; defaults to `False`.
            run (bool):
                Determines whether to run the generated sql or not; defaults to `None`
                which will reference the current value of the :attr:`auto_run` attribute
                which defaults to `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        # fmt: off
        try:
            schema, nm = p(nm)
            table = self._validate(
                val=(nm or self.nm), nm='nm', attr_nm='nm'
            )
            schema = self._validate(
                val=(schema or self.schema), nm='schema', attr_nm='schema'
            )
        except ValueError as e:
            raise e
        # fmt: on

        restrictions = {
            **(restrictions or dict()),
            **{
                "lower(table_name)": f"'{table.lower()}'",
                "lower(table_schema)": f"'{schema.lower()}'",
            },
        }
        if all_schemas:
            _ = restrictions.pop("lower(table_schema)")

        sql = self._info_schema_generic(
            obj="table", fields=fields, restrictions=restrictions, order_by=order_by
        )

        return self.sn.query(sql=sql) if self._run(run) else sql

    def info_schema_columns(
        self,
        nm: Optional[str] = None,
        fields: Optional[List] = None,
        restrictions: Optional[Dict] = None,
        order_by: Optional[List] = None,
        all_schemas: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Query ``information_schema.columns`` for a given table or view.

        Args:
            nm (str):
                Table name, including schema if creating a stage outside of the
                current schema.
            fields (List[str]):
                List of fields to include in returned results (e.g.
                ['ordinal_position', 'column_name', 'data_type'])
            restrictions (List[str]):
                List of conditionals typed as literal components of a `where`
                clause (e.g.["regexp_count(lower(column_name), 'tmstmp') = 0"]).
            order_by (List[str]):
                List of fields or their ordinal positions to order the results by.
            all_schemas (bool):
                Include tables/views from all schemas; defaults to `False`.
            run (bool):
                Determines whether to run the generated sql or not; defaults to `None`
                which will reference the current value of the :attr:`auto_run` attribute
                which defaults to `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        # fmt: off
        try:
            schema, nm = p(nm)
            table = self._validate(
                val=(nm or self.nm), nm='nm', attr_nm='nm'
            )
            schema = self._validate(
                val=(schema or self.schema), nm='schema', attr_nm='schema'
            )
        except ValueError as e:
            raise e
        # fmt: on
        restrictions = {
            **(restrictions or dict()),
            **{
                "lower(table_name)": f"'{table.lower()}'",
                "lower(table_schema)": f"'{schema.lower()}'",
            },
        }
        if all_schemas:
            restrictions.pop("lower(table_schema)")

        sql = self._info_schema_generic(
            obj="column", fields=fields, restrictions=restrictions, order_by=order_by
        )

        return self.sn.query(sql=sql) if self._run(run) else sql

    # TODO: (rename) 'cnt_records' --> 'cnt'
    def cnt_records(
            self,
            nm: Optional[str] = None,
            dst_of: Optional[str] = None,
            as_perc: Optional[bool] = None,
            run: Optional[bool] = None
    ):
        """Number of records within a table or view.

        Args:
            nm (str):
                Table name, including schema if creating a stage outside of the
                current schema.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        try:
            obj_name = self._validate(
                val=(nm or self.nm), nm="obj_name", attr_nm="obj_name"
            )
        except ValueError as e:
            raise e
        _from = f"from {obj_name}"
        sql = f"select count(*) {_from}"
        if dst_of:
            sql = f"select count(distinct {dst_of}) {_from}"
        if as_perc:
            sql = f"select count(distinct {dst_of}) / count(*) {_from}"
        return self.sn.query(sql=sql).snf.to_list(n=1) if self._run(run) else sql

    def table_last_altered(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Last altered timestamp for a table or view.

        Args:
            nm (str):
                Table name, including schema if creating a stage outside of the
                current schema.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        try:
            sql = self.info_schema_tables(
                nm=nm, fields=["table_name", "table_schema", "last_altered"]
            )
            return self.sn.query(sql=sql) if self._run(run) else sql
        except AssertionError as e:
            raise e

    def create_stage(
        self,
        nm_stage: str,
        nm_format: str,
        replace: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Create a staging table.

        Args:
            nm_stage (str):
                Name of stage to create, including schema if creating a stage
                outside of the current schema.
            nm_format (str):
                Name of file format to specify for the stage, including schema
                if using a format from outside of the current schema.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.
            replace (bool):
                Indicates whether to replace an existing stage if pre-existing;
                default is `False`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        create = self._create(replace=replace)
        _sql = f"{create} stage {nm_stage} file_format = {nm_format};"
        sql = strip(_sql)
        return self.sn.query(sql=sql) if self._run(run) else sql

    def drop(
        self,
        nm: Optional[str] = None,
        obj: Optional[str] = None,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Drop a ``Snowflake`` object.

        Args:
            nm (str):
                Name of the object to drop, including schema if creating a stage
                outside of the current schema.
            obj (str):
                Type of object to drop (e.g. 'table', 'schema', etc)
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        # fmt: off
        schema, nm = p(nm)
        try:
            obj_schema = self._validate(
                val=(schema or self.schema), nm='obj_schema', attr_nm='schema'
            )
            obj_name = self._validate(
                val=(nm or self.nm), nm='obj_name', attr_nm='obj_name'
            )
            obj = self._validate(
                val=(obj or self.obj), nm='obj', attr_nm='obj'
            )
        except ValueError as e:
            raise e
        # fmt: on
        _sql = (
            f"drop {obj} if exists {up(obj_schema)}.{up(obj_name)}"
            if obj.lower() in ["table", "view", "file_format"]
            else f"drop {obj} if exists {up(obj_name)}"
        )
        sql = strip(_sql)
        return self.sn.query(sql=sql) if self._run(run) else sql

    def clone(
        self,
        nm: Optional[str] = None,
        to: Optional[str] = None,
        obj: Optional[str] = None,
        run: Optional[bool] = None,
        replace: bool = False,
    ) -> Union[str, pd.DataFrame]:
        """Clone a ``Snowflake`` object.

        Warnings:
            *   Make sure to read `Snowflake's documentation
                <https://docs.snowflake.com/en/sql-reference/sql/create-clone.html>`_
                for restrictions and considerations when cloning objects.

        Note:
            *   In this specific method, the value provided to ``nm`` and ``to``
                can be a single object name, a single schema, or both in the
                form of `obj_schema.obj_name` depending on the desired outcome.
            *   Additionally, **at least one of the** ``nm`` **or** ``to``
                **arguments must be provided**.
            *   The defaults for the target object are constructed such that
                users can **either**:
                    1.  Clone objects to *other* schemas that inherit the
                        source object's *name* without specifying so in the
                        ``to`` argument, **or**
                    2.  Clone objects within the *current* schema that inherit
                        the source object's *schema* without specifying so in
                        the ``to`` argument.
            *   If providing a schema without a name to either argument, prefix
                the value provided with `__` to signify it's a schema and not
                a lower-level object to be cloned.
                    *   e.g. providing `nm='sample_table'` and
                        `to='__sandbox'` will clone `sample_table` from the
                        current schema to `sandbox.sample_table`.
            *   An assertion error will be raised raised if neither argument
                is specified as *this would result in a command to clone an
                object and store it in an object that has the same name &
                schema as the object being cloned*.

        Args:
            nm (str):
                Name of the object to clone, including schema if cloning an
                object outside of the current schema.
            to (str):
                Target name for cloned object, including schema if cloning an
                object outside of the current schema.
            obj (str):
                Type of object to clone (e.g. 'table', 'view', 'file-format');
                defaults to `table`.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.
            replace (bool):
                Indicates whether to replace an existing stage if pre-existing;
                default is `False`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        to_schema, to = p(nm=to)
        # fmt: off
        try:
            obj = self._validate(
                val=(obj or self.obj), nm='obj', attr_nm='obj'
            )
            schema = self._validate(
                val=(schema or self.schema), nm='schema', attr_nm='schema'
            )
            nm = self._validate(
                val=(nm or self.nm), nm='nm', attr_nm='nm'
            )

            to_schema = to_schema or schema
            to = to or nm
            if not to_schema and not to:
                raise ValueError(
                    "At least one of '__schema' or 'name` must be provided "
                    "in the 'to' argument of sql.clone()."
                )
            if nm == to and schema == to_schema:
                raise ValueError(
                    f"Target object name & schema mirrors source object name/schema. "
                    f"Please provide a different value `to`"
                )
        except ValueError as e:
            raise e
        # fmt: on
        create = self._create(replace=replace)
        _sql = (
            f"{create} {obj} {up(to_schema)}.{up(to)} " f"clone {up(schema)}.{up(nm)}"
        )
        sql = strip(_sql)
        return self.sn.query(sql=sql) if self._run(run) else sql

    def put_file_from_stage(
        self,
        path: Union[Path, str],
        nm_stage: str,
        options: Optional[Dict] = None,
        ignore_defaults: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Generates a 'put' command into a staging table from a local file.

        Args:
            path (Union[Path, str]):
                Path to local data file as a :class:`pathlib.Path` or string.
            nm_stage (str):
                Name of the staging table to load into.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.
            options (dict):
                Optional arguments to add to `put` statement in addition to
                the values specified in the ``loading.put`` section
                of **snowmobile.toml**.
            ignore_defaults (bool):
                Option to ignore the values specified in **snowmobile.toml**;
                defaults to `False`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        path = Path(str(path))
        statement = [f"put file://{path.as_posix()} @{nm_stage}"]
        # fmt: off
        defaults = (
            self.sn.cfg.loading.put.dict(by_alias=False) if not ignore_defaults
            else dict()
        )
        options = {
            **defaults,
            **(options or dict()),
        }
        for k, v in options.items():
            statement.append(f"\t{k} = {str(v).lower() if isinstance(v, bool) else v}")
        # fmt: on
        _sql = "\n".join(statement)
        sql = strip(_sql, trailing=False, whitespace=False, blanks=True)

        return self.sn.query(sql=sql) if self._run(run) else sql

    def copy_into_table_from_stage(
        self,
        nm: str,
        nm_stage: str,
        options: Optional[Dict] = None,
        ignore_defaults: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Generates a command to copy data into a table from a staging table.

        Args:
            nm (str):
                Name of the object to drop, including schema if creating a stage
                outside of the current schema.
            nm_stage (str):
                Name of the staging table to load from.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.
            options (dict):
                Optional arguments to add to `put` statement in addition to
                the values specified in the ``loading.put`` section
                of **snowmobile.toml**.
            ignore_defaults (bool):
                Option to ignore the values specified in **snowmobile.toml**;
                defaults to `False`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        statement = [f"copy into {nm} from @{nm_stage}"]
        defaults = (
            self.sn.cfg.loading.copy_into.dict(by_alias=False)
            if not ignore_defaults
            else dict()
        )
        options = {**defaults, **(options or dict())}
        for k, v in options.items():
            statement.append(f"\t{k} = {v}")
        _sql = "\n".join(statement)
        sql = strip(_sql, trailing=False, whitespace=False, blanks=True)
        return self.sn.query(sql=sql) if self._run(run) else sql

    def show_file_formats(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Lists all file formats in the current schema.

        Args:
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        sql = f"show file formats"
        return self.sn.query(sql=sql) if self._run(run) else sql

    def ddl(
        self,
        nm: Optional[str] = None,
        obj: Optional[str] = None,
        run: Optional[bool] = None,
    ) -> str:
        """Query the DDL for an in-warehouse object.

        Args:
            nm (str):
                Name of the object to get DDL for, including schema if object
                is outside of the current schema.
            obj (str):
                Type of object to get DDL for (e.g. 'table', 'view', 'file-format').
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (str):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        try:
            obj = self._validate(val=(obj or self.obj), nm="obj", attr_nm="obj")
            schema = self._validate(
                val=(schema or self.schema), nm="schema", attr_nm="schema"
            )
            nm = self._validate(val=(nm or self.nm), nm="nm", attr_nm="nm")
        except ValueError as e:
            raise e
        _sql = f"select get_ddl('{obj}', '{up(schema)}.{up(nm)}') as ddl"
        sql = strip(_sql)
        return self.sn.query(sql=sql).snf.to_list(n=1) if self._run(run) else sql

    def table_sample(
        self,
        nm: Optional[str] = None,
        n: Optional[int] = None,
        run: Optional[bool] = None,
    ) -> Union[str, pd.DataFrame]:
        """Select `n` sample records from a table.

        Args:
            nm (str):
                Name of table or view to sample, including schema if the table
                or view is outside of the current schema.
            n (int):
                Number of records to return, implemented as a 'limit' clause
                in the query; defaults to 1.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        # fmt: off
        try:
            schema = self._validate(
                val=(schema or self.schema), nm='schema', attr_nm='schema'
            )
            table = self._validate(
                val=(nm or self.nm), nm='nm', attr_nm='nm'
            )
        except ValueError as e:
            raise e
        # fmt: on
        limit = f"limit {n or 1}" if n != -1 else str()
        sql = f"""
select
    *
from {up(schema)}.{up(table)}
{limit}
        """
        return self.sn.query(sql=sql) if self._run(run) else sql

    def truncate(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Truncate a table.

        Args:
            nm (str):
                Name of table, including schema if the table is outside of the
                current schema.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        schema, nm = p(nm)
        # fmt: off
        try:
            schema = self._validate(
                val=(schema or self.schema), nm='schema', attr_nm='schema'
            )
            name = self._validate(
                val=(nm or self.nm), nm='nm', attr_nm='nm'
            )
        except ValueError as e:
            raise e
        # fmt: on
        _sql = f"truncate table {up(schema)}.{up(name)}"
        sql = strip(_sql)
        return self.sn.query(sql=sql) if self._run(run) else sql

    def current(
        self, obj: str, run: Optional[bool] = None
    ) -> Union[str, Union[str, int]]:
        """Generic implementation of 'select current' for session-based objects.

        Args:
            obj (str):
                Type of object to retrieve information for (schema, session, ..).
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        _sql = f"select current_{obj}()"
        sql = strip(_sql)
        return self.sn.query(sql=sql).snf.to_list(n=1) if self._run(run) else sql

    def current_session(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current session."""
        return self.current(obj="session", run=run)

    def current_schema(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current schema."""
        return self.current(obj="schema", run=run)

    def current_database(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current database."""
        return self.current(obj="database", run=run)

    def current_warehouse(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current warehouse."""
        return self.current(obj="warehouse", run=run)

    def current_role(self, run: Optional[bool] = None) -> Union[str, pd.DataFrame]:
        """Select the current role."""
        return self.current(obj="role", run=run)

    def use(self, nm: str, obj: str, run: Optional[bool] = None):
        """Generic implementation of 'use' command for in-warehouse objects.

        Args:
            nm (str):
                Name of object to use (schema name, warehouse name, role name, ..).
            obj (str):
                Type of object to use (schema, warehouse, role, ..).
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, pd.DataFrame]):
            Either:
                1.  The results of the query as a :class:`pandas.DataFrame`, or
                2.  The generated query as a :class:`str` of sql.

        """
        # fmt: off
        try:
            name = self._validate(
                val=(nm or self.nm), nm='nm', attr_nm='nm'
            )
        except ValueError as e:
            raise e
        # fmt: on
        _sql = f"use {obj} {up(name)}"
        sql = strip(_sql)
        return self.sn.query(sql=sql) if self._run(run) else sql

    def use_schema(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Use schema command."""
        return self.use(obj="schema", nm=nm, run=run)

    def use_database(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Use database command."""
        return self.use(obj="database", nm=nm, run=run)

    def use_warehouse(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Use warehouse command."""
        return self.use(obj="warehouse", nm=nm, run=run)

    def use_role(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, pd.DataFrame]:
        """Use role command."""
        return self.use(obj="role", nm=nm, run=run)

    def columns(
        self,
        nm: Optional[str] = None,
        from_info_schema: bool = False,
        run: Optional[bool] = None,
    ) -> Union[str, List]:
        """Returns an ordered list of columns for a table or view.

        note:
            *   The default behavior of this method is to retrieve the columns
                for a table or view by selecting a single sample record
                from the table and extracting the column names directly off
                the returned :class:`pandas.DataFrame` due to the performance
                gains in selecting a sample record as opposed to querying the
                ``information_schema.columns``.
            *   This can be changed by passing `from_info_schema=False`.

        Args:
            nm (str):
                Name of table or view, including schema if the table or view is
                outside of the current schema.
            from_info_schema (bool):
                Indicates whether to retrieve columns via the
                ``information_schema.columns`` or by selecting a sample record
                from the table or view; defaults to `False`.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, List]):
            Either:
                1.  An ordered list of columns for the table or view, **or**
                2.  The query against the table or view as a :class:`str` of sql.

        """
        if from_info_schema:
            return self._columns_from_info_schema(nm=nm, run=run)
        else:
            return self._columns_from_sample(nm=nm, run=run)

    # noinspection PyBroadException
    def exists(self, nm: Optional[str] = None) -> bool:
        """Checks the existence of a table or view.

        Args:
            nm (str):
                Name of table or view, including schema if the table or view is
                outside of the current schema.

        Returns (bool):
            Boolean indication of whether or not the table or view exists.

        """
        try:
            _ = self.table_sample(nm=nm)
            return True
        except:
            return False

    def _columns_from_info_schema(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, List]:
        """Retrieves list of columns for a table or view **from information schema**.

        Args:
            nm (str):
                Name of table or view, including schema if the table or view is
                outside of the current schema.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, List]):
            Either:
                1.  An ordered list of columns for the table or view, **or**
                2.  The query against ``information_schema.columns`` as a
                    :class:`str` of sql.

        """
        sql = self.info_schema_columns(
            nm=nm, fields=["ordinal_position", "column_name"], order_by=[1], run=False
        )
        return self.sn.query(sql).snf.to_list(col="column_name") if run else sql

    def _columns_from_sample(
        self, nm: Optional[str] = None, run: Optional[bool] = None
    ) -> Union[str, List]:
        """Retrieves a list of columns for a table or view from **sampling table**.

        Args:
            nm (str):
                Name of table or view, including schema if the table or view is
                outside of the current schema.
            run (bool):
                Indicates whether to execute generated sql or return as string;
                default is `True`.

        Returns (Union[str, List]):
            Either:
                1.  An ordered list of columns for the table or view, **or**
                2.  The query against the table or view as a :class:`str` of sql.

        """
        _sql = self.table_sample(nm=nm, run=False, n=1)
        sql = strip(_sql)
        return list(self.sn.query(sql, lower=False).columns) if self._run(run) else sql

    @staticmethod
    def _create(replace: bool = False):
        """Utility to generate 'create'/'create or replace' based on an argument."""
        return "create" if not replace else "create or replace"

    def _run(self, run: Union[bool, None]) -> bool:
        """Determines whether or not to execute a piece of sql.

        Used in all subsequent methods containing a `run` argument.

        Args:
            run (Union[bool, None]):
                The value from a method's `run` argument.

        Returns (bool):
            Boolean value indicating whether or not to execute the sql generated by
            the method to which the value of `run` was passed.

        note:
            *   The default value of `run` in all subsequent methods is ``None``.
            *   When any method of :class:`SQL` containing a `run` argument is called,
                the argument's value is passed to this method which returns either:
                    1.  The argument's value if it is a valid bool (i.e. a user-provided
                        value to the method), or
                    2.  The boolean representation of the current :attr:`auto_run`
                        attribute (`True` by default).

        """
        if isinstance(run, bool):
            return run
        else:
            return bool(self.auto_run)

    def _info_schema_generic(
        self,
        obj: str,
        fields: List[str] = None,
        restrictions: Dict[str, str] = None,
        order_by: Optional[List] = None,
    ) -> str:
        """Generic case of selecting from information schema tables/columns.

        Queries different parts of the information schema based on an ``obj``
        and the mapping of object type to information schema defined in
        `snowmobile.core.sql._map_information_schema.py`.

        """
        info_schema_loc = self.sn.cfg.sql.info_schema_loc(obj=obj)
        fields = self.fields(fields=fields)
        where = self.where(restrictions=restrictions)
        order_by = self.order(by=order_by)

        sql = f"""
select
{fields}
from {info_schema_loc}
{where}
{order_by}
"""
        return strip(sql, trailing=False, blanks=True)

    @staticmethod
    def order(by: List[Union[int, str]]) -> str:
        """Generates 'order by' clause from a list of fields or field ordinal positions."""
        if by:
            order_by_fields = ",".join(str(v) for v in by)
            return f"order by {order_by_fields}"
        else:
            return str()

    @staticmethod
    def where(restrictions: Dict) -> str:
        """Generates a 'where' clause based on a dictionary of restrictions.

        Args:
            restrictions (dict):
                A dictionary of conditionals where each key/value pair
                respectively represents the left/right side of a condition
                within a 'where' clause.

        Returns (str):
            Formatted where clause.

        """
        if restrictions:
            args = [
                f"{str(where_this)} = {str(equals_this)}"
                for where_this, equals_this in restrictions.items()
            ]
            args = "\n\tand ".join(args)
            return f"where\n\t{args}"
        else:
            return str()

    @staticmethod
    def fields(fields: List) -> str:
        """Utility to generate fields within a 'select' statement."""
        return "\n".join(
            f'\t{"," if i > 1 else ""}{f}'
            for i, f in enumerate(fields or ["*"], start=1)
        )

    @staticmethod
    def _validate(
        val: Optional[str, int], nm: str, attr_nm: Optional[str] = None
    ) -> str:
        """Validates the value of an argument passed to a method.

        This method is built to validate method arguments in instances where an
        unspecified argument can fall back to an attribute if it has been set.

        Each of the 'closing' variables below represents a different ending to
        a sentence within the exception message depending on the value provided
        from the method and if the attribute the argument falls back to has been
        set at the time the method is called.

        Args:
            val (Union[str, int]:
                Value to validate.
            nm (str):
                Name of argument in the method being called.
            attr_nm (str):
                Name of attribute to fall back to if the boolean representation
                of ``val`` is `False`.

        """
        if not val:
            closing1 = (
                "." if not attr_nm else f", nor is its fallback attribute '{attr_nm}'."
            )
            closing2 = (
                "."
                if not attr_nm
                else f" or set the '{attr_nm}' attribute before calling the method."
            )
            raise ValueError(
                f"\nValue provided for '{nm}' is not valid{closing1}\n"
                f"Please provide a valid value for '{nm}'{closing2}"
            )
        return val

    def _reset(self):
        self.schema = self.sn.cfg.connection.current.schema_name
        self.nm = None
        self.obj = "table"
        return self

    def copy(self) -> SQL:
        """User-facing copy method."""
        return self.__copy__()

    def __copy__(self) -> SQL:
        """Dunder copy method."""
        return type(self)(sn=self.sn, nm=f"{self.schema}.{self.nm}", obj=self.obj)

    def __call__(self, *args, **kwargs) -> SQL:
        for k, v in kwargs.items():
            if k in vars(self):
                setattr(self, k, v)
        return self

    def __str__(self) -> str:
        return f"snowmobile.SQL(creds='{self.sn.cfg.connection.creds}')"

    def __repr__(self) -> str:
        return f"snowmobile.SQL(creds='{self.sn.cfg.connection.creds}')"



"""
Default values for local file paths used by ``snowmobile``.
"""
from pathlib import Path

# ====================================
HERE = Path(__file__).absolute()
DIR_MODULES = HERE.parent
DIR_PKG_DATA = DIR_MODULES / "pkg_data"

EXTENSIONS_DEFAULT_PATH = DIR_PKG_DATA / "snowmobile_ext.toml"
DDL_DEFAULT_PATH = DIR_PKG_DATA / "DDL.sql"
SQL_EXPORT_HEADING_DEFAULT_PATH = DIR_PKG_DATA / "sql_export_heading.txt"
# ====================================



"""
Demonstrate specifying an alternate snowmobile.toml file *path*.
../docs/snippets/snowmobile/specifying_configuration.py
"""
import snowmobile
from pathlib import Path

path = Path.cwd() / 'snowmobile_v2.toml'  # any alternate file path

sn = snowmobile.connect(from_config=path)
# snowmobile-include



"""
Instantiate a snowmobile.Configuration object directly.
../docs/snippets/instantiate_configuration.py
"""

import snowmobile

cfg = snowmobile.Configuration()

type(cfg)  # > snowmobile.core.configuration.Configuration
vars(cfg)



"""
[external-sources] from **snowmobile.toml**.
"""
from __future__ import annotations

from pathlib import Path

from pydantic import Field

from .base import Base


class Location(Base):
    """[external-sources]"""

    # fmt: off
    ddl: Path = Field(
        default_factory=Path, alias="ddl"
    )
    extensions: Path = Field(
        default_factory=Path, alias="snowmobile_ext"
    )
    sql_export_heading: Path = Field(
        default_factory=Path, alias="sql-export-heading"
    )
    # fmt: on



"""
Instantiate `script` from 'overview.sql' and demonstrate 'nm' resolution.
../docs/snippets/script/overview-statement-names.py
"""

# Setup -----------------------------------------------------------------------
from pathlib import Path
paths = {p.name: p for p in Path.cwd().glob('**/*.sql')}
path = paths['overview.sql']

import snowmobile

sn = snowmobile.connect(delay=True)

# Example ---------------------------------------------------------------------

# ==== nm ====
script = snowmobile.Script(path=path, sn=sn)
script.dtl()

# Store statements 1 and 4 for inspection
s1, s4 = script(1), script(4)

print(s1.nm)      #> create table~s1
print(s4.nm)      #> select all~sample_table

print(s1.desc)    #> s1
print(s4.desc)    #> sample_table

print(s1.anchor)  #> create table
print(s4.anchor)  #> select all

print(sn.cfg.script.patterns.core.delimiter)  #> ~

# -- Block 2
print(s4.anchor_ge)  #> select data
print(s4.anchor_pr)  #> select all
print(s4.anchor)     #> select all

print(s4.desc_ge)    #> s4
print(s4.desc_pr)    #> sample_table
print(s4.desc)       #> sample_table

print(s4.nm_ge)      #> select data~s4
print(s4.nm_pr)      #> select all~sample_table
print(s4.nm)         #> select all~sample_table

# -- Definitions --

print(s1.nm)      #> create table~s1
print(s4.nm)      #> select all~sample_table

print(s1.desc)    #> s1
print(s4.desc)    #> sample_table

print(s1.anchor)  #> create table
print(s4.anchor)  #> select all

print(s1.kw)  #> create
print(s4.kw)  #> select



"""
This file is to store objects accessed by the entire test suite within the scope
of:
    *   Constant variables
    *   A base test class and an associated `idfn` for pytest IDs
    *   Static utility functions

"""
from pathlib import Path

from pydantic import BaseModel


# -- CONSTANT VARIABLES ---

CREDS = "snowmobile_testing"
CONFIG_FILE_NM = "snowmobile_testing.toml"

TESTS_ROOT = Path(__file__).absolute().parent
FILES = {p.name: p for p in TESTS_ROOT.rglob("*") if p.is_file()}


# -- BASE TEST CLASS ------


class BaseTest(BaseModel):
    """Base object for snowmobile test classes.

    Note:
        *   This class is intended to be used as a **namespace only** and
            should not be extended with additional methods unless necessary
            in order to maintain test readability.
        *   Its extensions should define a __repr__ function that will be
            retrieved by pytest when :func:`tests.idfn()` is called on its instances.
        *   Built from pydantic's BaseModel as opposed to a straight dataclass
            since it will enforce type safety, enables annotating attributes
            with clear descriptions while referencing them in code with
            succinct variable names, and is already a project dependency.

    """

    @property
    def pytest_id(self) -> str:
        """Pytest ID for console feedback during tests runs."""
        return self.__repr__()

    def __repr__(self) -> str:
        """Valid __repr__ string to fully reproduce the object under test."""
        pass

    class Config:
        """Enabling so that arbitrary types can be set on derived instances."""

        arbitrary_types_allowed = True


def idfn(val: BaseTest):
    """Retrieves :attr:`pytest_id` from derived instances of `BaseTest`."""
    assert hasattr(val, "pytest_id"), f"No 'pytest_id' for class {type(val)}"
    return val.pytest_id


# -- UTILITY FUNCTIONS ---


def get_validation_file(path1: Path, sub_dir: str = "expected_outcomes") -> Path:
    """Returns a path for the validation file given a test file path."""
    export_dir = path1.parent.parent  # expected: '.snowmobile.
    offset = path1.relative_to(export_dir)
    return export_dir / sub_dir / offset


def contents_are_identical(path1: Path, path2: Path) -> bool:
    """Reads in two paths as text files and confirms their contents are identical."""
    with open(path1, "r") as r1:
        f1 = r1.read()
    with open(path2, "r") as r2:
        f2 = r2.read()
    return f1 == f2


def script(script_name: str):
    """Get a script object from its script name."""
    from snowmobile import Script, Snowmobile

    path_to_script = FILES[script_name]
    return Script(
        path=path_to_script,
        sn=Snowmobile(config_file_nm=CONFIG_FILE_NM, creds=CREDS, delay=True),
    )




# =============================================================================
# ========= Configuration file for the Sphinx documentation builder ===========
# =============================================================================


# -- Import Management --------------------------------------------------------
import os
import sys
from typing import List

from pathlib import Path

sys.path.append(".")

HERE = Path(__file__).absolute()
ROOT = HERE.parent.parent
DOCS_DIR = ROOT / "docs"
LINKS_DIR = DOCS_DIR / "links"
EXT_DIR = DOCS_DIR / "ext"
PACKAGE_DIR = ROOT / "snowmobile"

to_insert = {
    "DOCS_DIR": DOCS_DIR,
    # "LINKS_DIR": LINKS_DIR,
    "EXT_DIR": EXT_DIR,
    "PACKAGE_DIR": PACKAGE_DIR,
    "ROOT": ROOT,
}
for dir_name, dir_path in to_insert.items():
    sys.path.insert(0, str(dir_path))
    print(f"<added> {dir_name}: {dir_path.as_posix()}")

os.chdir(str(ROOT))

# -- Runtime extensions -------------------------------------------------------

# ==== xref ====
# -- See docstring in ext/xref.py for more info
# noinspection PyUnresolvedReferences
from links.link import *

# noinspection PyUnresolvedReferences
from links import *

# ==== snippets ====
from ext import snippets
snippets.write()

# ==== MySt AutoAPI ====
# from ext import autoapi_myst

# -- Standard Options ---------------------------------------------------------

source_suffix = {
    '.md': 'myst-nb',
    '.rst': 'restructuredtext',
    '.rst.txt': 'restructuredtext',
    '.ipynb': 'myst-nb',
    '.myst': 'myst-nb',
}

extensions = [
    "myst_nb",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "autoapi.extension",
    "xref",
    "sphinx_panels",
    "sphinx_copybutton",
    "sphinx_togglebutton",
    # "myst_parser",
    # "nbsphinx",
]

master_doc = "index"
default_role = None
language = "python"
templates_path = ["_templates"]
exclude_patterns = ["_build", "__main__.py", "__init__.py", "**.ipynb_checkpoints"]
autodoc_typehints = 'description'


# Project Information ---------------------------------------------------------

# TODO: Sort out the requirements dependencies; see Build #12815403
from snowmobile import __version__

# __version__ = "0.2.0"
# version = __version__

project = "snowmobile"
copyright = "2020, Grant E Murray"
author = "Grant E Murray"
version = __version__
release = __version__
# release = version


# Sphinx Panels ---------------------------------------------------------------

# https://sphinx-panels.readthedocs.io/en/latest/

html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.7.0/css/font-awesome.min.css"
]
panels_add_bootstrap_css = True  # required for color in icons/etc
panels_add_fontawesome_latex = True

panels_css_variables = {
    "tabs-color-label-active": "rgba(33, 150, 243, 0.85)",
    "tabs-color-label-inactive": "rgba(33, 150, 243, 0.55)",
    "tabs-color-overline": "#329ef452",
    # "tabs-color-underline": "#329ef452",
    "tabs-color-underline": "#ffffff",
    "tabs-size-label": "0.75rem",
}

# MySt ------------------------------------------------------------------------

# https://myst-parser.readthedocs.io/en/latest/

myst_heading_anchors = 5  # auto generate anchor slugs for h1-h5
autosectionlabel_max_depth = 4
autosectionlabel_prefix_document = True

from markdown_it.extensions import deflist

myst_enable_extensions = [
    # "dollarmath",
    # "amsmath",
    # "deflist",
    # "html_image",
    # "colon_fence",
    # "smartquotes",
    # "replacements",
    # "linkify",
    # "substitution",
    "amsmath",
    "colon_fence",
    "deflist",
    "dollarmath",
    "html_image",
]
myst_url_schemes = ("http", "https", "mailto")


# MySt Nb ---------------------------------------------------------------------
# jupyter_execute_notebooks = "cache"
jupyter_execute_notebooks = "off"

kernelspec = {
    'display_name': 'Snowmobile', 'language': 'python', 'name': 'snowmobile2'
}

nb_render_priority = {
  "html": (
            "application/vnd.jupyter.widget-view+json",
            "application/javascript",
            "text/html",
            "image/svg+xml",
            "image/png",
            "image/jpeg",
            "text/markdown",
            "text/latex",
            "text/plain",
        )
}
# default; see https://myst-nb.readthedocs.io/en/latest/use/formatting_outputs.html

# -- CopyButton ---------------------------------------------------------------

# https://sphinx-copybutton.readthedocs.io/en/latest/

# Throwing out outputs, keeping jupyter inputs and code blocks.
# copybutton_selector = "div:not(.output) > div.highlight pre, div.notranslate td.code div.highlight pre"
copybutton_selector = "div:not(.output) div:not(.sn-output) > div.highlight pre, div.notranslate td.code div.highlight pre"
# copybutton_selector = "div.highlight pre"  # default


# -- AutoAPI ------------------------------------------------------------------

# https://sphinx-autoapi.readthedocs.io/en/latest/reference/config.html

# autoapi_add_toctree_entry = False
autoapi_add_toctree_entry = True
autoapi_type = "python"
autoapi_dirs = ["../snowmobile"]
autoapi_ignore = [
    "__main__.py",
    "__init__.py",
    "**/stdout/*",
    "**_runner/*",
    "**/.snowmobile/*",
]
autoapi_python_class_content = "class"  # 'both' if __init__ as well
autoapi_member_order = "bysource"  # 'bysource' or 'groupwise'
# 'bysource' means 'by docstring order' for attributes and 'by actual source'
# for methods and properties.


# -- ToggleButton -------------------------------------------------------------

# https://sphinx-togglebutton.readthedocs.io/en/latest/

togglebutton_hint = "show"


# Sphinx Material / HTML Theme ------------------------------------------------

# https://bashtage.github.io/sphinx-material/

import sphinx_material

html_title = "snowmobile"
html_show_sourcelink = True
html_sidebars = {"**": ["globaltoc.html", "localtoc.html", "searchbox.html"]}
html_theme_path = sphinx_material.html_theme_path()
html_context = sphinx_material.get_html_context()
html_theme = "sphinx_material"
extensions.append("sphinx_material")

html_theme_options = {
    # 'google_analytics_account': 'UA-XXXXX',
    # 'logo_icon': '&#xe869',
    "html_minify": True,
    "html_prettify": False,
    "css_minify": True,
    # "globaltoc_depth": 3,
    "globaltoc_depth": 1,
    "globaltoc_collapse": True,
    "globaltoc_includehidden": True,
    "repo_url": "https://github.com/GEM7318/Snowmobile",
    "repo_name": "gem7318/snowmobile",
    "nav_title": "",
    "color_primary": "blue",
    "color_accent": "cyan",
    # "color_accent": "orange",
    # "color_primary": "lightcyan",
    # "theme_color": "2196f3",
    "theme_color": "3391c6",
    "body_max_width": None,
    'master_doc': True,
    'version_dropdown': True,

    'nav_links': [
        {
            'href': './usage/snowmobile.html/connecting-to-snowflake',
            'title': 'Connecting to Snowflake',
            'internal': True,
        },
    ]
}
# Accent colors:
# red, pink, purple, deep-purple, indigo, blue, light-blue, cyan,
# teal, green, light-green, lime, yellow, amber, orange, deep-orange


# Other HTML Options ----------------------------------------------------------

html_static_path = ["_static"]
html_use_smartypants = True
html_use_index = True
html_domain_indices = True
html_show_sphinx = False
html_show_copyright = False

# pygments_style = "sphinx"
pygments_style = "emacs"

# html_sidebars = {}
# html_additional_pages = {}
# html_domain_indices = True
# html_split_index = False
# html_last_updated_fmt = '%b %d, %Y'


# Napoleon --------------------------------------------------------------------

napoleon_google_docstring = True
napoleon_numpy_docstring = False

napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = True

napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False

napoleon_use_param = True
napoleon_use_ivar = False
napoleon_use_rtype = False
napoleon_attr_annotations = True
napoleon_use_keyword = True

napoleon_custom_sections = "Attributes"

# napoleon_type_aliases = {
#     "DictCursor": "SnowflakeCursor",
# }


# -- External URLs ------------------------------------------------------------

python_version = ".".join(map(str, sys.version_info[0:2]))

intersphinx_mapping = {
    "sqlparse": ("https://sqlparse.readthedocs.io/en/latest/", None),
    "pandas": ("http://pandas.pydata.org/pandas-docs/stable", None),
    "sphinx": ("http://www.sphinx-doc.org/en/stable", None),
    "python": ("https://docs.python.org/" + python_version, None),
    "matplotlib": ("https://matplotlib.org", None),
    "numpy": ("https://docs.scipy.org/doc/numpy", None),
    "sklearn": ("http://scikit-learn.org/stable", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
}


rm_dirs = [
    r'_build', r'.ipynb_checkpoints', r'jupyter_execute'
]

html_js_files = [
    'js/scrollbar.js',
    'js/ios.js',
    'js/resize.js',
]


# TODO: Exclude all caps too
def autoapi_skip_member(app, what, name, obj, skip, options):
    """Exclude all private attributes, methods, and dunder methods from Sphinx."""
    import re
    exclude = re.findall("\._.*", str(obj)) or "stdout" in str(obj).lower()
    return skip or exclude


def rm_build_dirs(to_rm: List[str] = None, root_dir: Path = DOCS_DIR):
    """Ensure fresh build each time."""
    import shutil
    to_rm = to_rm or rm_dirs
    dirs = [d for d in root_dir.iterdir() if d.is_dir() and d.name in to_rm]
    for d in dirs:
        shutil.rmtree(path=d, ignore_errors=True)


def setup(app):
    """Add autoapi-skip-member."""

    # -- setup
    if not os.environ.get('READTHEDOCS'):
        # rm_build_dirs()
        try:
            rm_build_dirs()
        except Exception as e:
            print(e)

    # -- api docs
    app.connect("autoapi-skip-member", autoapi_skip_member)

    # -- theme
    app.add_css_file('css/friendly.css')         # additional template
    app.add_css_file('css/application_ext.css')  # snowmobile

    # -- dynamic header
    # app.add_js_file("js/scrollbar.js")
    # autoapi_myst.to_myst()




"""
Light extensions to pd.DataFrame object.
"""

import datetime
import itertools
import re
from typing import Dict, List, Optional, Tuple

import pandas as pd

from . import errors
from .column import Column
from . import Generic


@pd.api.extensions.register_dataframe_accessor("snf")
class SnowFrame(Generic):
    """Light extension to :class:`pandas.DataFrame` with a ``snf`` entry point.

    """

    def __init__(self, df: pd.DataFrame):
        super().__init__()
        self._obj: pd.DataFrame = df
        self._cols: List[Column] = [Column(c) for c in self._obj.columns]
        self.cols: List[Column] = [Column(c) for c in self._obj.columns]

    def shared_cols(self, df2: pd.DataFrame) -> List[Tuple[pd.Series, pd.Series]]:
        """Returns list of tuples containing column pairs that are common between two DataFrames.
        """
        for col in set(self._obj.columns) & set(df2.columns):
            yield self._obj[col], df2[col]

    @staticmethod
    def series_max_diff_abs(col1: pd.Series, col2: pd.Series, tolerance: float) -> bool:
        """Determines if the max **absolute** difference between two
        :class:`pandas.Series` is within a tolerance level.

        """
        try:
            diff = (col1.astype(float) - col2.astype(float)).abs().max()
            return diff <= abs(tolerance)
        except TypeError as e:
            raise TypeError(e)

    @staticmethod
    def series_max_diff_rel(col1: pd.Series, col2: pd.Series, tolerance: float) -> bool:
        """Determines if the maximum **relative** difference between two
        :class:`pandas.Series` is within a tolerance level.

        """
        try:
            diff = ((col1.astype(float) / col2.astype(float)) - 1).abs().max()
            return diff <= abs(tolerance)
        except TypeError as e:
            raise TypeError(e)

    def df_max_diff_abs(self, df2: pd.DataFrame, tolerance: float) -> bool:
        """Determines if the maximum **absolute** difference between any value
        in the shared columns of 2 DataFrames is within a tolerance level.

        """
        return all(
            self.series_max_diff_abs(col1=c[0], col2=c[1], tolerance=tolerance)
            for c in self.shared_cols(df2=df2)
        )

    def df_max_diff_rel(self, df2: pd.DataFrame, tolerance: float) -> bool:
        """Determines if the maximum **relative** difference between any value
        in the shared columns of 2 DataFrames is within a tolerance level.

        """
        return all(
            self.series_max_diff_rel(col1=c[0], col2=c[1], tolerance=tolerance)
            for c in self.shared_cols(df2=df2)
        )

    def df_diff(
        self,
        df2: pd.DataFrame,
        abs_tol: Optional[float] = None,
        rel_tol: Optional[float] = None,
    ) -> bool:
        """Determines if the column-wise difference between two DataFrames is
        within a relative **or** absolute tolerance level.

        note:
            *   ``df1`` and ``df2`` are assumed to have a shared, pre-defined index.
            *   Exactly **one** of ``abs_tol`` and ``rel_tol`` is expected to be a
                a valid float; the other is expected to be **None**.
            *   If valid float values are provided for both ``abs_tol`` and ``rel_tol``,
                the outcome of the maximum **absolute** difference with respect to
                ``abs_tol`` will be returned regardless of the value of ``rel_tol``.

        Args:
            df2 (pd.DataFrame): 2nd DataFrame for comparison.
            abs_tol (float): Absolute tolerance; default is None.
            rel_tol (float): Relative tolerance; default is None.

        Returns (bool):
            Boolean indicating whether or not difference is within tolerance.

        """
        if isinstance(abs_tol, float):
            return self.df_max_diff_abs(df2=df2, tolerance=abs_tol)
        else:
            return self.df_max_diff_rel(df2=df2, tolerance=rel_tol)

    def partitions(self, on: str) -> Dict[str, pd.DataFrame]:
        """Returns a dictionary of DataFrames given a DataFrame and a partition column.

        Note:
            *   The number of distinct values within ``partition_on`` column will be
                1:1 with the number of partitions that are returned.
            *   The ``partition_on`` column is dropped from the partitions that are returned.
            *   The depth of a vertical concatenation of all partitions should equal the
                depth of the original DataFrame.

        Args:
            on (str):
                The column name to use for partitioning the data.

        Returns (Dict[str, pd.DataFrame]):
            Dictionary of {(str) partition_value: (pd.DataFrame) associated subset of df}

        """
        partitioned_by = set(self._obj[on])
        if len(partitioned_by) < 2:
            raise errors.SnowFrameInternalError(
                msg=(
                    f"Found one distinct value, '{list(partitioned_by)[0]}' "
                    f"within '{on}' column of DataFrame. A minimum of 2 is required."
                )
            )
        base_partitions = {p: self._obj[self._obj[on] == p] for p in partitioned_by}
        return {p: df.drop(columns=[on]) for p, df in base_partitions.items()}

    # noinspection PyUnresolvedReferences
    def ddl(self, table: str) -> str:
        """Returns a string containing 'create table' DDL given a table name"""
        return (
            pd.io.sql.get_schema(self._obj, table)
            .replace("CREATE TABLE", "CREATE OR REPLACE TABLE")
            .replace('"', "")
        )

    def lower(self, col: Optional[str] = None) -> pd.DataFrame:
        """Lower cases all column names **or** all values within `col` if provided."""
        if col:
            self._obj[col] = self._obj[col].apply(lambda x: str(x).lower())
        else:
            self._obj.rename(
                columns={c.prior: c.lower() for c in self.cols}, inplace=True
            )
        return self._obj

    def upper(self, col: Optional[str] = None) -> pd.DataFrame:
        """Upper cases all column names **or** all values within `col` if provided."""
        if col:
            self._obj[col] = self._obj[col].apply(lambda x: str(x).lower())
        else:
            self._obj.rename(
                columns={c.prior: c.upper() for c in self.cols}, inplace=True
            )
        return self._obj

    def reformat(self):
        """Re-formats DataFrame's columns via :class:`Column.reformat()`."""
        self._obj.rename(
            columns={c.prior: c.reformat() for c in self.cols}, inplace=True
        )
        return self._obj

    def append_dupe_suffix(self):
        """Adds a trailing index number '_i' to duplicate column names."""
        cols = pd.Series(self._obj.columns)
        for dupe_col in cols[cols.duplicated()].unique():
            dupe_indices = cols[cols == dupe_col].index.values.tolist()
            cols[dupe_indices] = [
                f"{dupe_col}_{i}" if i != 0 else dupe_col
                for i in range(sum(cols == dupe_col))
            ]
        self._obj.columns = cols

    def to_list(self, col: Optional[str] = None, n: Optional[int] = None) -> List:
        """Succinctly retrieves a column as a list.

        Args:
            col (str):
                Name of column.
            n (int):
                Number of records to return; defaults to full depth of column.

        """
        if not col:
            col = list(self._obj.columns)[0]
        as_list = list(self._obj[col])
        return as_list if not n else as_list[n - 1]

    def add_tmstmp(self, col_nm: Optional[str] = None) -> pd.DataFrame:
        """Adds a column containing the current timestamp to a DataFrame.

        Args:
            col_nm (str):
                Name for column; defaults to `LOADED_TMSTMP`.

        """
        col = Column(original=(col_nm or "LOADED_TMSTMP"), src="snowmobile")
        self.cols.append(col)
        self._obj[col.current] = datetime.datetime.now()
        return self._obj

    @property
    def original(self) -> pd.DataFrame:
        """Returns the DataFrame in its original form (drops columns added by
        :class:`SnowFrame` and reverts to original column names).

        """
        df: pd.DataFrame = self._obj[
            [c.current for c in self.cols if c.src == "original"]
        ]
        return df.rename(columns={c.current: c.original for c in self.cols})

    @property
    def has_dupes(self) -> bool:
        """DataFrame has duplicate column names."""
        return len(set(self._obj.columns)) != len(list(self._obj.columns))

    def cols_matching(
        self, patterns: List[str], ignore_patterns: List[str] = None
    ) -> List[str]:
        """Returns a list of columns given a list of patterns to find.

        Args:
            patterns (List[str]):
                List of regex patterns to match columns on.
            ignore_patterns (List[str]):
                Optional list of regex patterns to exclude.

        Returns (List[str]):
            List of columns found/excluded.

        """
        to_ignore = list(
            itertools.chain.from_iterable(
                p if isinstance(p, list) else [p] for p in (ignore_patterns or [])
            )
        )
        return [
            col
            for col in self._obj.columns
            if any(re.findall(p_incl, col) for p_incl in patterns)
            and not any(re.findall(col, p_excl) for p_excl in to_ignore)
        ]

    def cols_ending(self, nm: str, ignore_patterns: Optional[List] = None) -> List[str]:
        """Returns all columns up to ``nm`` in a DataFrame.

        Args:
            nm (str):
                Name of column to end index at.
            ignore_patterns (List[str]):
                Optional list of regex patterns to exclude in the list that's
                returned; primarily used to for getting `end-index-at` list
                while excluding `src_description`.

        Returns (List[str]):
            List of column names matching criterion.

        """
        matches = []
        ignore_patterns = ignore_patterns or []
        for i, col in enumerate(self._obj.columns, start=1):
            if col == nm:
                matches = [
                    c
                    for c in list(self._obj.columns)[:i]
                    if not any(list(re.findall(p, c)) for p in ignore_patterns)
                ]
        return matches



# -*- coding: utf-8 -*-

import pytest

collect_ignore = ["__init__.py", "pkg_data", "_stdout.py"]

from tests import CONFIG_FILE_NM, CREDS


@pytest.fixture(scope="session")
def sn():
    """Returns a standard `Connector` object."""
    import snowmobile

    sn = snowmobile.connect(creds=CREDS, config_file_nm=CONFIG_FILE_NM)
    sn.ex("create schema TESTING_SNOWMOBILE")
    yield sn
    sn.ex("drop schema TESTING_SNOWMOBILE")


@pytest.fixture(scope="session")
def sn_delayed():
    """Returns a delayed `Connector` object."""
    import snowmobile

    return snowmobile.connect(creds=CREDS, config_file_nm=CONFIG_FILE_NM, delay=True)


@pytest.fixture(scope="session")
def sql(sn_delayed):
    """Returns a sql object; connection omitted."""
    from snowmobile import SQL

    return SQL(sn=sn_delayed, auto_run=True)



"""
Verify `default-creds` has been changed to `creds2`.
../docs/snippets/snowmobile/verify_default_alias_change.py
"""
import snowmobile

sn = snowmobile.connect()

assert sn.cfg.connection.default_alias == 'creds2', (
    "Something's not right here; expected default_alias =='creds2'"
)
# snowmobile-include



"""
Export template `snowmobile.toml` to a specified directory.
../docs/snippets/export_config.py
"""
import snowmobile
from pathlib import Path

snowmobile.Configuration(export_dir=Path.cwd())

# -- complete example; should run 'as is' --



"""
Establish an initial connection and explore :class:`Connector` attributes.
../docs/snippets/test_connection.py
"""
import snowmobile

sn = snowmobile.connect()  # optionally provide `creds='credentials_alias'`
assert sn.alive

type(sn)  # > snowmobile.core.connection.Snowmobile

type(sn.cfg)  # > snowmobile.core.configuration.Configuration
str(sn.cfg)  # > snowmobile.Configuration('snowmobile.toml')

type(sn.con)  # > snowflake.connector.connection.SnowflakeConnection
type(sn.cursor)  # > snowflake.connector.cursor.SnowflakeCursor

df1 = sn.query("select 1")  #  == pd.read_sql()
type(df1)  # > pandas.core.frame.DataFrame

cur1 = sn.query("select 1", as_df=False)  #  == SnowflakeConnection.cursor().execute()
type(cur1)  # > snowflake.connector.cursor.SnowflakeCursor

import pandas as pd

df2 = pd.read_sql(sql="select 1", con=sn.con)
cur2 = sn.con.cursor().execute("select 1")

print(df2.equals(df1))  # > True
print(cur1.fetchone() == cur2.fetchone())  # > True

# -- complete example; should run 'as is' --



"""Tests for snowmobile.Snowmobile."""
import pytest

import snowmobile

from tests import CONFIG_FILE_NM, CREDS


@pytest.mark.connector
def test_basic_query(sn):
    """Verify a standard connector object connects to the DB."""
    df = sn.query("select 1")
    assert not df.empty, "expected df.empty=False"


@pytest.mark.connector
def test_basic_query_via_ex(sn):
    """Verify a standard connector object connects to the DB."""
    cur = sn.ex("select 1")
    assert cur.fetchone()[0] == 1


@pytest.mark.connector
def test_query_into_cursor(sn):
    """Verifies passing `as_cur=True` to snowmobile.query() returns a cursor."""
    from snowflake.connector.connection import SnowflakeCursor

    cur = sn.query("select 1", as_cur=True)
    assert isinstance(cur, SnowflakeCursor)


@pytest.mark.connector
def test_query_into_dictcursor(sn):
    """Verifies passing `as_dcur=True` to snowmobile.query() returns a dictcursor."""
    from snowflake.connector import DictCursor

    dcur = sn.query("select 1", as_dcur=True)
    assert isinstance(dcur, DictCursor)


@pytest.mark.connector
def test_query_as_scalar(sn):
    """Verifies passing `as_dcur=True` to snowmobile.query() returns a dictcursor."""
    scalar = sn.query("select 1", as_scalar=True)
    assert scalar == 1


@pytest.mark.connector
def test_alive_evaluates_to_false_on_delayed_connection(sn_delayed):
    """Verify a delayed connector object does not connect to the DB."""
    assert not sn_delayed.alive


@pytest.mark.connector
def test_cursor_is_accessible_from_delayed_connection(sn_delayed):
    """Verify a delayed connector object does not connect to the DB."""
    assert not sn_delayed.cursor.is_closed()


@pytest.mark.connector
def test_alive_evaluates_to_false_post_disconnect(sn):
    """Verifies connector.disconnect() closes session."""
    assert not sn.disconnect().alive, "expected sn_delayed.alive=False post-disconnect"


# noinspection PyUnresolvedReferences
@pytest.mark.connector
def test_alternate_kwarg_takes_precedent_over_configuration_file():
    """Tests over-riding configuration file with alternate connection kwargs."""
    sn_as_from_config = snowmobile.connect(creds=CREDS, config_file_nm=CONFIG_FILE_NM)
    sn_with_a_conflicting_parameter = snowmobile.connect(
        creds=CREDS,
        config_file_nm=CONFIG_FILE_NM,
        autocommit=False,  # <-- explicit kwarg that also exists in snowmobile.toml
    )

    assert (
        # verify `config.autocommit=True`
        sn_as_from_config.con._autocommit
        # verify `autocommit=False` kwarg took precedent over config
        and not sn_with_a_conflicting_parameter.con._autocommit
    )


@pytest.mark.connector
def test_providing_invalid_credentials_raises_exception(sn):
    """Verify an invalid set of credentials raises DatabaseError."""
    from snowflake.connector.errors import DatabaseError

    with pytest.raises(DatabaseError):
        snowmobile.connect(
            creds=CREDS,
            config_file_nm=CONFIG_FILE_NM,
            user="invalid@invalid.com",  # <-- a set of invalid credentials
        )


# noinspection SqlResolve
@pytest.mark.connector
def test_invalid_sql_passed_to_query_raises_exception(sn):
    """Tests that invalid sql passed to snowmobile.query() raises DatabaseError."""
    from pandas.io.sql import DatabaseError

    with pytest.raises(DatabaseError):
        sn.query("select * from *")  # <-- an invalid sql statement


# noinspection SqlResolve
@pytest.mark.connector
def test_invalid_sql_passed_to_ex_raises_exception(sn):
    """Tests that invalid sql passed to connector.ex() raises ProgrammingError."""
    from snowflake.connector.errors import ProgrammingError

    with pytest.raises(ProgrammingError):
        sn.ex("select * from *")  # <-- an invalid sql statement


# noinspection SqlResolve
@pytest.mark.connector
def test_invalid_sql_passed_to_exd_raises_exception(sn):
    """Tests that invalid sql passed to connector.ex() raises ProgrammingError."""
    from snowflake.connector.errors import ProgrammingError

    with pytest.raises(ProgrammingError):
        sn.exd("select * from *")  # <-- an invalid sql statement


@pytest.mark.connector
def test_invalid_credentials_alias_raises_exception():
    """Tests that an invalid credentials name creds raises KeyError."""
    from snowmobile import Snowmobile

    with pytest.raises(KeyError):
        Snowmobile(creds="name_for_a_nonexistent_set_of_creds")


@pytest.mark.connector
def test_masked_dunder_str_method_for_sets_of_credentials(sn):
    """Verifies that the __str()__ method for a set of credentials is masked."""
    assert all(
        not line.split("=")[-1].strip("'").replace("*", "")
        for line in str(sn.cfg.connection.credentials).split("\n")
        if str(line).startswith("  ")
    )


@pytest.mark.connector
def test_dunder_repr_is_valid(sn):
    assert sn.__repr__()


@pytest.mark.connector
def test_invalid_arguments_provided_to_query_raises_value_error(sn):
    """Verifies an error gets thrown if two return types are specified."""
    with pytest.raises(ValueError):
        sn.query('select 1', as_df=True, as_dcur=True)



"""
Base class for all module-level Stdout objects.
"""
from __future__ import annotations

from pathlib import Path
from typing import Optional, Union


class Console:
    def __init__(self):
        pass

    @staticmethod
    def offset_path(
        file_path: Union[str, Path],
        root_dir_nm: Optional[str] = None,
        indent: Optional[str] = None,
        output: bool = False,
    ) -> str:
        """Gets truncated string for file path relative to a root directory.

        Args:
            file_path (Union[str, Path]):
                Full string or Path to the file.
            root_dir_nm (str):
                Directory name to anchor truncated file path to.
            indent (str):
                Optional character to indent by (e.g. '\t')
            output (bool):
                Option to print the string as well as return it.

        Returns (str):
            Truncation of a path for console output.

        """
        indent = indent or str()
        root_dir = root_dir_nm or file_path.parent.name
        root_path = file_path.parent
        for p in file_path.parents:
            if p.name == root_dir:
                root_path = p
                break
        offset_path = file_path.relative_to(root_path)
        stdout = f"{indent}../{root_path.name}/{offset_path.as_posix()}"
        if output:
            print(stdout)
        return stdout



"""
.
../snippets/intro_connector.py
"""
import snowmobile

sn = snowmobile.connect()
sn2 = snowmobile.connect(creds="creds1")

sn.cfg.connection.current == sn2.cfg.connection.current  #> True
sn.sql.current("schema") == sn2.sql.current("schema")    #> True
sn.sql.current("session") == sn2.sql.current("session")  #> False

type(sn)         #> snowmobile.core.connection.Snowmobile

type(sn.cfg)     #> snowmobile.core.configuration.Configuration
str(sn.cfg)      #> snowmobile.Configuration('snowmobile.toml')

type(sn.con)     #> snowflake.connector.connection.SnowflakeConnection
type(sn.cursor)  #> snowflake.connector.cursor.SnowflakeCursor

df1 = sn.query("select 1")  #  == pd.read_sql()
type(df1)                   #> pandas.core.frame.DataFrame

cur1 = sn.ex("select 1")    #  == SnowflakeConnection.cursor().execute()
type(cur1)                  #> snowflake.connector.cursor.SnowflakeCursor

dcur1 = sn.exd("select 1")  #  == SnowflakeConnection.cursor(DictCursor).execute()
type(dcur1)                 #> snowflake.connector.DictCursor


import pandas as pd

df2 = pd.read_sql(sql="select 1", con=sn.con)

cur2 = sn.con.cursor().execute("select 1")

from snowflake.connector import DictCursor

dcur2 = sn.con.cursor(cursor_class=DictCursor).execute("select 1")

print(df2.equals(df1))                       #> True
print(cur1.fetchone() == cur2.fetchone())    #> True
print(dcur1.fetchone() == dcur2.fetchone())  #> True

cur1 = sn.cursor.execute("select 1")
cur2 = sn.cursor.execute("select 2")

cursor = sn.cursor
cur11 = cursor.execute("select 1")
cur22 = cursor.execute("select 2")

id(cur1) == id(cur2)    #> False
id(cur11) == id(cur22)  #> True

# -- complete example; should run 'as is' --



"""
Module for post-processing attributes of ``snowmobile.Script`` in conjunction
with configuration options stored in *snowmobile.toml*.

These result in two files being exported into a `.snowmobile` folder in the
same directory as the .sql file that ``snowmobile.Script`` was instantiated
with.

Header-levels and formatting of tagged information is configured in the
*snowmobile.toml* file, with defaults resulting in the following structure::

        ```md

        # Script Name.sql         *[script name gets an 'h1' header]
        ----

        - **Tag1**: Value1         [tags are bolded, associated values are not]
        - **Tag2**: Value2         [same for all tags/attributes found]
        - ...

        **Description**           *[except for the 'Description' section]
                                  *[this is just a blank canvas of markdown..]
                                  *[..but this is configurable]

        ## (1) create-table~dummy_name *[contents get 'h2' level headers]
        ----

        - **Tag1**: Value1       *[tags can also be validations arguments..
        - **Arg1**: Val2          [that snowmobile will run on the sql results]

        **Description**          *[contents get one of these too]

        **SQL**                  *[their rendered sql does as well]
            ...sql
                ...
                ...
            ...


        ## (2) update-table~dummy_name2
        ----
        [structure repeats for all contents in the script]

        ```

"""
from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from . import Generic, Configuration, Snowmobile, Diff, Empty, Section, Statement
from .paths import DIR_PKG_DATA
from .cfg import Marker
from .utils import Console


class Markup(Generic):
    """Markup document from SQL script.
    """

    def __init__(
        self,
        sn: Snowmobile,
        path: Path,
        contents: Dict[int, Union[Statement, Marker]],
        alt_file_nm: Optional[str] = None,
        alt_file_prefix: Optional[str] = None,
        alt_file_suffix: Optional[str] = None,
        incl_sql: bool = True,
        incl_markers: bool = True,
        incl_raw: bool = False,
        sql_incl_export_disclaimer: bool = True,
    ):
        super().__init__()
        self._stdout = Console()
        self.pkg_data_dir = DIR_PKG_DATA
        self.contents = contents
        self.cfg: Configuration = sn.cfg
        self.alt_file_nm: str = alt_file_nm or str()
        self.alt_file_prefix: str = alt_file_prefix or str()
        self.alt_file_suffix: str = alt_file_suffix or str()
        self.sql_incl_export_disclaimer: bool = sql_incl_export_disclaimer
        self.incl_markers: bool = incl_markers
        self.incl_sql = incl_sql
        self.incl_raw = incl_raw
        self.exported: List[Path] = list()
        self.created: List[Path] = list()
        self.path = Path(str(path)) if path else Path()

    @property
    def doc_root(self) -> Path:
        """Documentation sub-directory; `.snowmobile` by default."""
        return (
            # Path()
            # if not self.path.anchor
            # else
            self.path.parent
            / self.cfg.script.export_dir_nm
        )

    @property
    def _file_nm(self) -> str:
        """Generic file name of script."""
        return self.alt_file_nm or self.path.name

    @property
    def _file_nm_components(self) -> Tuple[str, str]:
        """Utility for easy access to the stem and the extension of file name."""
        stem, _, ext = self._file_nm.rpartition(".")
        return stem, ext

    @property
    def file_nm_sql(self) -> str:
        """Adjusted file name of the exported sql script."""
        stem, ext = self._file_nm_components
        return f"{self.alt_file_prefix}{stem}{self.alt_file_suffix}.{ext}"

    @property
    def file_nm_md(self) -> str:
        """Adjusted file name of the exported markdown."""
        stem, ext = self._file_nm_components
        return f"{self.alt_file_prefix}{stem}{self.alt_file_suffix}.md"

    @property
    def script_dir(self) -> Path:
        """Directory for all exports from specific _file_nm.."""
        stem, _, ext = self._file_nm.rpartition(".")
        return self.doc_root / stem

    @property
    def path_md(self) -> Path:
        """Full path to write markdown to."""
        return self.script_dir / self.file_nm_md

    @property
    def path_sql(self) -> Path:
        """Full path to write sql """
        return self.script_dir / self.file_nm_sql

    @property
    def sections(self) -> Dict[int, Section]:
        """All header sections of markdown file as a dictionary."""
        sections = {}
        for i, s in self.contents.items():
            if self._is_statement(s=s):  # create section from statement.section()
                sections[i] = s.as_section(self.incl_raw)
            else:
                sections[i] = Section(  # create section from marker metadata
                    incl_raw=self.incl_raw,
                    is_multiline=True,
                    cfg=self.cfg,
                    raw=s.raw,
                    **s.as_args(),
                )

        return {i: sections[i] for i in sorted(sections)}

    @property
    def markdown(self) -> str:
        """Full markdown file as a string."""
        included = self.included
        return "\n\n".join(s.md for i, s in self.sections.items() if i in included)

    @property
    def _export_disclaimer(self) -> str:
        """Block comment disclaimer of export at top of exported sql file."""
        path_to_sql_txt = self.pkg_data_dir / "sql_export_heading.txt"
        with open(path_to_sql_txt, "r") as r:
            header = r.read()
        return f"{header}" if self.sql_incl_export_disclaimer else str()

    @property
    def included(self):
        """All included indices based on incl_ attributes."""
        return {
            i
            for i, s in self.contents.items()
            if (
                (self.incl_sql and self._is_statement(s=s))
                or (self.incl_markers and not self._is_statement(s=s))
            )
        }

    @property
    def sql(self):
        """SQL for export."""
        to_export = [
            s.trim()
            if self._is_statement(s)
            else self.cfg.script.as_parsable(raw=s.raw, is_marker=True)
            for i, s in self.contents.items()
            if i in self.included
        ]
        if self.sql_incl_export_disclaimer:
            to_export.insert(0, self._export_disclaimer)
        return "\n".join(to_export)

    @staticmethod
    def _is_statement(s: Union[Statement, Diff, Empty, Marker]) -> bool:
        """Utility to check if a given instance of contents is a statement."""
        return isinstance(s, (Statement, Diff, Empty))

    def _scaffolding(self) -> None:
        """Ensures directory scaffolding exists before attempting export."""
        if not self.script_dir.exists():
            self.script_dir.mkdir(parents=True)
            self.created.append(self.script_dir)

    def _export(self, path: Path, val: str):
        """Ensure directory scaffolding exists and writes a string to a path (.sql or .md)."""
        self._scaffolding()
        with open(path, "w") as f:
            f.write(val)
            self.exported.append(path)
            self._stdout.offset_path(
                file_path=path, root_dir_nm=path.parent.name, indent="\t", output=True
            )

    def export(self, md_only: bool = False, sql_only: bool = False) -> None:
        """Export files.

        Args:
            md_only (bool): Export markdown file only.
            sql_only (bool): Export sql file only.

        """

        def _md():
            self._export(path=self.path_md, val=self.markdown)

        def _sql():
            self._export(path=self.path_sql, val=self.sql)

        if md_only:
            _md()
        elif sql_only:
            _sql()
        else:
            _md()
            _sql()

    def __call__(self, **kwargs):
        """Batch setattr function for all keywords matching Markup's attributes."""
        return self.cfg.batch_set_attrs(obj=self, attrs=kwargs)

    def __str__(self) -> str:
        return f"script.Markup('{self.file_nm_sql}')"

    def __repr__(self) -> str:
        return f"script.Markup('{self.file_nm_sql}')"



"""Test loader"""
import pytest

from tests import CONFIG_FILE_NM, CREDS

TESTING_TABLE_NAME = 'test_snowmobile_table'


@pytest.fixture(scope='session')
def rnp():
    """Returns numpy reference with the seed set to 9."""
    import numpy as np

    def rnp(s: int = 1):
        """Sets numpy seed and returns library reference."""
        np.random.seed(s)
        return np
    return rnp


@pytest.fixture(scope='session')
def df(rnp):
    """Sets seed and returns a DataFrame to use for tests."""
    import pandas as pd

    return pd.DataFrame(
        {
            'col1': [rnp().random.choice(['spam', 'eggs', 'ham']) for _ in range(24)],
            'col2': [rnp().random.choice(['alpha', 'beta', 'gamma']) for _ in range(24)],
            'col3': [i for i in range(24)],
        }
    )


@pytest.mark.table
def test_table(sn, df):
    """Temporary - verify the most basic operations of a loader object."""
    import snowmobile
    table_name = "test_snowmobile_upload"
    table = snowmobile.Table(df=df, table=table_name, sn=sn)
    loaded = table.load(if_exists="replace")
    assert loaded


@pytest.mark.table
def test_table_core_functionality(sn, df, rnp):
    """Kitchen sink testing of core snowmobile.Table functionality"""
    import snowmobile
    import pandas as pd

    # Ex. 1 -------------------------------------------------------------------

    sn.sql.exists(TESTING_TABLE_NAME)

    t1 = snowmobile.Table(sn=sn, table=TESTING_TABLE_NAME, df=df, as_is=True)
    assert t1.loaded

    # Ex. 11 ------------------------------------------------------------------

    from snowmobile.core.errors import ExistingTableError

    with pytest.raises(ExistingTableError):
        snowmobile.Table(
            sn=sn,
            table=TESTING_TABLE_NAME,
            df=df,
            as_is=True,
            if_exists='fail',
        )

    # Ex. 2 -------------------------------------------------------------------

    df2 = pd.concat([df, df], axis=1)

    from snowmobile.core.errors import ColumnMismatchError

    t2 = snowmobile.Table(sn=sn, table=TESTING_TABLE_NAME, df=df2)

    with pytest.raises(ColumnMismatchError):
        t2.load()

    # Ex. 22
    t2.load(if_exists='replace')
    assert t2.loaded

    # Ex. 3 -------------------------------------------------------------------
    with pytest.raises(ValueError):
        t2.load(if_exists='some_invalid_if_exists_value')



"""
Base class for all :class:`Statement` objects.
"""
from __future__ import annotations

import time
from typing import Any, Callable, Dict, Optional, Union, Tuple

import pandas as pd
import sqlparse
from IPython.display import Markdown, display
from pandas.io.sql import DatabaseError as pdDataBaseError
from snowflake.connector.errors import DatabaseError, ProgrammingError

from . import ExceptionHandler, Section, Name, errors, cfg
from . import Generic, Snowmobile  # isort: skip


class Statement(Name, Generic):
    """Base class for all :class:`Statement` objects.

    Home for attributes and methods that are associated with **all** statement
    objects, generic or QA.

    Attributes:
        sn (snowmobile.connect):
            :class:`snowmobile.connect` object.
        statement (Union[sqlparse.sql.Statement, str]):
            A :class:`sqlparse.sql.Statement` object.
        index (int):
            The context-specific index position of a statement within a script;
            can be `None`.
        _index (int):
            The original index position of a statement as read from the source;
            This is used to restore the index position of a given statement
            before exiting a specified context within :class:`snowmobile.Script`.
        patterns (config.Pattern):
            :class:`config.Pattern` object for more succinct access to
            values specified in **snowmobile.toml**.
        results (pd.DataFrame):
            The results of the statement if executed as a :class:`pandas.DataFrame`.
        outcome (int):
            Numeric indicator of outcome; defaults to `0` and is modified
            based on the outcome of statement execution and/or QA validation
            for derived classes.
        outcome_txt (str):
            Plain text of outcome ('skipped', 'failed', 'completed', 'passed').
        outcome_html (str):
            HTML text for the outcome as an admonition/information banner
            based on the following mapping of :attr:`outcome_txt` to
            admonition argument:
                *   `failed` ------> `warning`
                *   `completed` --> `info`
                *   `passed` -----> `success`
        start_time (int):
            Unix timestamp of the query start time if executed; 0 otherwise.
        end_time (int):
            Unix timestamp of the query end time if executed; 0 otherwise.
        execution_time (int):
            Execution time of the query in seconds if executed; 0 otherwise.
        execution_time_txt (str):
            Plain text description of execution time if executed; returned in
            seconds if execution time is less than 60 seconds, minutes otherwise.
        attrs_raw (str):
            A raw string of the tag/attributes associated with the statement.
        attrs_parsed (dict):
            A parsed dictionary of the tag/attributes associated with the statement.
        is_tagged (bool):
            Indicates whether or not the statement is tagged by the user.
        is_multiline (bool):
            Indicates whether or not a statement tag is a multiline tag; will
            be `False` by default if :attr:`is_tagged` is `False`.
        first_keyword (sqlparse.sql.Token):
            The first keyword within the statement as a :class:`sqlparse.sql.Token`.
        sql (str):
            The sql associated with the statement as a raw string.
        tag (Name):
            Statement's :class:`~snowmobile.core.name.Name`.

    """

    # fmt: off
    _PROCESS_OUTCOMES: Dict[Any, Any] = {
        -3: ("success", "passed"),
        -2: ("warning", "failed"),
        -1: ("-", "error: post-processing"),
        # -- derived end
        0: ("-", ""),
        # -- base start
        1: ("-", "error: execution"),
        2: ("success", "completed"),
    }

    _DERIVED_FAILURE_MAPPING = {
        'qa-diff': errors.QADiffFailure,
        'qa-empty': errors.QAEmptyFailure
    }
    # fmt: on

    def __init__(
        self,
        sn: Snowmobile,
        statement: Union[sqlparse.sql.Statement, str],
        index: Optional[int] = None,
        attrs_raw: Optional[str] = None,
        e: Optional[ExceptionHandler] = None,
        **kwargs,
    ):
        Generic.__init__(self)

        self._index: int = index
        self._outcome: int = int()

        self.outcome: bool = True
        self.executed: bool = bool()

        self.sn = sn
        self.statement: sqlparse.sql.Statement = sn.cfg.script.ensure_sqlparse(
            statement
        )

        self.patterns: cfg.Pattern = sn.cfg.script.patterns
        self.results: pd.DataFrame = pd.DataFrame()

        self.start_time: int = int()
        self.end_time: int = int()
        self.execution_time: int = int()
        self.execution_time_txt: str = str()

        self.attrs_raw = attrs_raw or str()
        self.is_multiline = "\n" in self.attrs_raw
        self.is_tagged: bool = bool(self.attrs_raw)
        self._sql = sn.cfg.script.isolate_sql(s=self.statement)
        self.attrs_parsed, nm = self.parse()

        Name.__init__(
            self, index=index, sql=self.sql, nm_pr=nm, configuration=self.sn.cfg
        )

        self.e = e or ExceptionHandler(within=self)

    @property
    def sql(self):
        """Raw sql from statement, including result limit if enabled."""
        if self.sn.cfg.script.result_limit in [-1, 0]:
            return self._sql
        return f"{self._sql}\nlimit {self.sn.cfg.script.result_limit}"

    def parse(self) -> Tuple[Dict, str]:
        """Parses a statement tag into a valid dictionary.

        Uses the values specified in **snowmobile.toml** to parse a
        raw string of statement arguments into a valid dictionary.

        note:
            *   If :attr:`is_multiline` is `True` and `name` is not included
                within the arguments, an assertion error will be thrown.
            *   If :attr:`is_multiline` is `False`, the raw string within
                the tag will be treated as the name.
            *   The :attr:`tag` attribute is set once parsing is completed
                and name has been validated.

        Returns (dict):
            Parsed tag arguments as a dictionary.

        """
        if not self.is_tagged:
            return dict(), str()

        if self.is_multiline:
            attrs_parsed = self.sn.cfg.script.parse_str(block=self.attrs_raw)
            if "name" in attrs_parsed:
                name = attrs_parsed.pop("name")
            else:
                try:
                    name = self.sn.cfg.script.parse_name(raw=self.attrs_raw)
                except errors.InvalidTagsError as e:
                    raise e

            return attrs_parsed, name

        return dict(), self.attrs_raw

    def start(self):
        """Sets :attr:`start_time` attribute."""
        self.start_time = time.time()

    def end(self):
        """Updates execution time attributes.

        In total, sets:
            *   :attr:`end_time`
            *   :attr:`execution_time`
            *   :attr:`execution_time_txt`

        """
        self._outcome, self.outcome, self.executed = 2, True, True
        self.end_time = time.time()
        self.execution_time = int(self.end_time - self.start_time)
        self.execution_time_txt = (
            f"{self.execution_time}s"
            if self.execution_time < 60
            else f"{int(self.execution_time/60)}m"
        )

    @property
    def attrs_total(self) -> Dict:
        """Parses namespace for attributes specified in **snowmobile.toml**.

        Searches attributes for those matching the keys specified in
        ``script.markdown.attributes.aliases`` within **snowmobile.toml**
        and adds to the existing attributes stored in :attr:`attrs_parsed`
        before returning.

        Returns (dict):
            Combined dictionary of statement attributes from those explicitly
            provided within the script and from object's namespace if specified
            in **snowmobile.toml**.

        """
        current_namespace = {
            **self.sn.cfg.attrs_from_obj(
                obj=self, within=list(self.sn.cfg.attrs.from_namespace)
            ),
            **self.sn.cfg.methods_from_obj(
                obj=self, within=list(self.sn.cfg.attrs.from_namespace)
            ),
        }
        namespace_overlap_with_config = set(
            current_namespace
        ).intersection(  # all from current namespace
            self.sn.cfg.attrs.from_namespace
        )  # snowmobile.toml
        attrs = {k: v for k, v in self.attrs_parsed.items()}  # parsed from .sql
        for k in namespace_overlap_with_config:
            attr = current_namespace[k]
            attr_value = attr() if isinstance(attr, Callable) else attr
            if attr_value:
                attrs[k] = attr_value
        return attrs

    def trim(self) -> str:
        """Statement as a string including only the sql and a single-line tag name.

        note:
            The tag name used here will be the user-provided tag from the
            original script or a generated :attr:`Name.nm` if a tag was not
            provided for a given statement.

        """
        patterns = self.sn.cfg.script.patterns
        open_p, close_p = patterns.core.to_open, patterns.core.to_close
        return f"{open_p}{self.nm}{close_p}\n{self.sql};\n"

    def render(self) -> Statement:
        """Renders the statement's sql as markdown in Notebook/IPython environments."""
        display((Markdown(self.as_section().sql_md)))
        return self

    @property
    def is_derived(self):
        """Indicates whether or not it's a generic or derived (QA) statement."""
        return self.anchor in self.sn.cfg.QA_ANCHORS

    @property
    def lines(self) -> int:
        """Depth of the statement's sql by number of lines."""
        return len(self.sql.split("\n"))

    def as_section(self, incl_raw: Optional[bool] = None) -> Section:
        """Returns current statement as a :class:`Section` object."""
        return Section(
            index=self.index,
            h_contents=self.nm,
            parsed=self.attrs_total,
            raw=self.attrs_raw,
            sql=self.sql,
            cfg=self.sn.cfg,
            results=self.results,
            incl_raw=incl_raw,
            is_multiline=self.is_multiline,
        )

    def set_state(
        self,
        index: Optional[int] = None,
        ctx_id: Optional[int] = None,
        in_context: Optional[bool] = None,
        filters: dict = None,
    ) -> Statement:
        """Sets current state/context on a statement object.

        Args:
            ctx_id (int):
                Unix timestamp the :meth:`script.filter()` context manager was
                invoked.
            filters (dict):
                Kwargs passed to :meth:`script.filter()`.
            index (int):
                Integer to set as the statement's index position.

        """
        if index:
            self.index = index
        if ctx_id:
            self.e.set(ctx_id=ctx_id)
        if isinstance(in_context, bool):
            self.e.set(in_context=in_context)
        if filters:
            super().scope(**filters)
        return self

    def reset(
        self,
        index: bool = False,
        ctx_id: bool = False,
        in_context: bool = False,
        scope: bool = False,
    ) -> Statement:
        """Resets attributes on the statement object to reflect as if read from source.

        In its current form, includes:
            *   Resetting the statement/tag's index to their original values.
            *   Resetting the :attr:`is_included` attribute of the statement's
                :attr:`tag` to `True`.
            *   Populating :attr:`error_last` with errors from current context.
            *   Caching current context's timestamp and resetting back to `None`.

        """
        if index:
            self.index = self._index
        if ctx_id:
            self.e.reset(ctx_id=True)
        if in_context:
            self.e.reset(in_context=True)
        if scope:
            super().scope(**{})
        return self

    def process(self):
        """Used by derived classes for post-processing the returned results."""
        return self

    def run(
        self,
        as_df: bool = True,
        lower: bool = True,
        render: bool = False,
        on_error: Optional[str] = None,
        on_exception: Optional[str] = None,
        on_failure: Optional[str] = None,
        ctx_id: Optional[int] = None,
    ) -> Statement:
        """Run method for all statement objects.

        Args:
            as_df (bool):
                Store results of query as :class:`pandas.DataFrame` or
                :class:`SnowflakeCursor`.
            lower (bool):
                Lower case column names in :attr:`results` DataFrame if
                `as_df=True`.
            render (bool):
                Render the sql executed as markdown.
            on_error (str):
                Behavior if an execution/database error is encountered
                    * `None`: default behavior, exception will be raised
                    * `c`: continue with execution
            on_exception (str):
                Behavior if an exception is raised in the **post-processing**
                of results from a derived class of :class:`Statement` (
                :class:`Empty` and :class:`Diff`).
                    * `None`: default behavior, exception will be raised
                    * `c`: continue with execution
            on_failure (str):
                Behavior if no error is encountered in execution or post-processing
                but the result of the post-processing has turned the statement's
                :attr:`outcome` attribute to False, indicating the results
                returned by the statement have failed validation.
                    * `None`: default behavior, exception will be raised
                    * `c`: continue with execution

        Returns (Statement):
            Statement object post-executing query.

        """
        self.e.set(ctx_id=(ctx_id or -1))
        try:
            if self:
                self.start()
                self.results = self.sn.query(self.sql, as_df=as_df, lower=lower)
                self.end()
                self.e.set(outcome=2)
        except (ProgrammingError, pdDataBaseError, DatabaseError) as e:
            self.e.collect(e=e).set(outcome=1)
        finally:
            # only post-process when execution did not raise database error
            if self.e.outcome != 1:
                self.process()
        # fmt: off
        # ---------------------------
        if (
            self.e.seen(  # db error raised during execution
                of_type=errors.db_errors, to_raise=True
            )
            and on_error != "c"  # stop on execution error
        ):
            raise self.e.get(
                of_type=errors.db_errors,
                to_raise=True,
                first=True,
            )
        # ---------------------------
        if (
            self.e.seen(  # post-processing error occurred
                of_type=errors.StatementPostProcessingError, to_raise=True
            )
            and on_exception != "c"  # stop on post-processing exception
        ):
            raise self.e.get(
                of_type=errors.StatementPostProcessingError,
                to_raise=True,
                first=True,
            )
        # ---------------------------
        if (
            self.is_derived  # is child class with `.process()` method
            and not self.outcome  # outcome of `.process()` did not pass
            and on_failure != "c"  # stop on failure of `.process()`
        ):
            raise self.e.get(
                of_type=list(self._DERIVED_FAILURE_MAPPING.values()),
                to_raise=True,
                first=True,
            )
        # ---------------------------
        # fmt: on

        if render:
            self.render()

        return self

    def set_sql(self, sql: str) -> None:
        """Public entry point to manually set the statement's .sql attribute."""
        self._sql = sql

    @staticmethod
    def _validate_parsed(attrs_parsed: Dict):
        """Returns args to verify 'name' attribute is present in a multiline tag."""
        condition, msg = (
            attrs_parsed.get("name"),
            f"Required attribute 'name' not found in multi-line tag's "
            f"arguments;\n attributes found are: {','.join(list(attrs_parsed))}",
        )
        return condition, msg

    def outcome_txt(self, _id: Optional[int] = None) -> str:
        """Outcome as a string."""
        return self._PROCESS_OUTCOMES[_id or self.e.outcome or 0][1]

    # TODO: Move this to patterns
    @property
    def outcome_html(self) -> str:
        """Outcome as an html admonition banner."""
        alert = self._PROCESS_OUTCOMES[self.e.outcome or 0][0]
        return f"""
<div class="alert-{alert}">
<center><b>====/ {self.outcome_txt()} /====</b></center>
</div>""".strip()

    def __bool__(self):
        """Determined by the value of :attr:`Name.is_included`."""
        return self.is_included

    def __str__(self) -> str:
        return f"Statement('{self.nm}')"

    def __repr__(self) -> str:
        return f"Statement('{self.nm}')"



"""
External links extensions for sphinx.
https://sublime-and-sphinx-guide.readthedocs.io/en/latest/references.html
"""

# from os.path import dirname, basename, isfile
#
# from pathlib import Path
# current_dir = Path.cwd().absolute()
#
# # modules = [f for f in current_dir.iterdir() if f.is_file()]
# modules = [basename(f)[:-3] for f in current_dir.iterdir() if f.is_file()]
# __all__ = modules
# print(__all__)

__name__ = "links"

from os.path import dirname, basename, isfile

import glob

modules = glob.glob(dirname(__file__) + "/*.py")

__all__ = [basename(f)[:-3] for f in modules if isfile(f)]



"""
Unit tests for :class:`ExceptionHandler`.
"""

import pytest

import time
from typing import Union, Tuple, Type

from snowmobile.core.exception_handler import ExceptionHandler

from snowmobile.core.errors import (
    StatementNotFoundError,
    DuplicateTagError,
    QAFailure,
    QAEmptyFailure,
    QADiffFailure,
    InternalError,
    Error,
)

errors = Union[
    StatementNotFoundError,
    DuplicateTagError,
    QAFailure,
    QAEmptyFailure,
    QADiffFailure,
    InternalError,
    Error,
]


@pytest.fixture(scope="session")
def default_test_exception_handler() -> ExceptionHandler:
    """ExceptionHandler with completely default values."""
    return ExceptionHandler()


@pytest.fixture(scope="session")
def snowmobile_test_exception_handler(sn_delayed) -> ExceptionHandler:
    """ExceptionHandler with completely default values."""
    from snowmobile.core.exception_handler import ExceptionHandler

    return ExceptionHandler(within=sn_delayed)


@pytest.fixture(scope="session")
def any_exception_marked_not_to_be_raised1() -> Tuple[errors, Type[errors]]:
    """sample1"""
    e = InternalError(nm="1st example exception", msg="testing 1st")
    return e, InternalError


@pytest.fixture(scope="session")
def any_exception_marked_not_to_be_raised2() -> Tuple[errors, Type[errors]]:
    """sample2"""
    e = StatementNotFoundError(nm="2nd example exception", msg="testing 2nd")
    return e, StatementNotFoundError


@pytest.fixture(scope="session")
def to_be_raised_exception1() -> Tuple[errors, Type[errors]]:
    """FIRST SAMPLE EXCEPTION MARKED AS TO BE RAISED."""
    e = DuplicateTagError(nm="3rd example exception", msg="testing 3rd", to_raise=True)
    return e, DuplicateTagError


@pytest.fixture(scope="session")
def to_be_raised_exception2() -> Tuple[errors, Type[errors]]:
    """SECOND SAMPLE EXCEPTION MARKED AS TO BE RAISED."""
    e = QAFailure(nm="4th example exception", msg="testing 4th", to_raise=True, idx=0)
    return e, QAFailure


@pytest.fixture(scope="session")
def any_unique_integer1() -> int:
    """To be used as `ctx_id` #1."""
    return time.time_ns()


@pytest.fixture(scope="session")
def any_unique_integer2() -> int:
    """To be used as `ctx_id` #2."""
    time.sleep(1)
    return time.time_ns()



"""
Instantiate `script` from 'overview.sql' and inspect high-level contents.
../docs/snippets/script/overview-base-parsing.py
"""

# Setup -----------------------------------------------------------------------
from pathlib import Path
paths = {p.name: p for p in Path.cwd().glob('**/*.sql')}
path = paths['overview.sql']

import snowmobile

sn = snowmobile.connect(delay=True)

# Example ---------------------------------------------------------------------

# -- Block 1 --
script = snowmobile.Script(path=path, sn=sn)
script.dtl()
"""
>>>
overview.sql
============
1: Statement('create table~s1')
2: Statement('insert into~s2')
3: Statement('select data~s3')
4: Statement('select all~sample_table')
5: Statement('create transient table~s5')
6: Statement('insert into~s6')
7: Statement('drop table~s7')
"""

# -- Block 2 --
# Store a few statements, accessed by index position
s_first, s_last = script(1), script(-1)

# first sql keyword
print(s_first.kw)  #> create
print(s_last.kw)   #> drop

# position within `script`
print(s_first.index)  #> 1
print(s_last.index)   #> 7

# -- Block 3 --
script.run(1)    # .run() from `script`
script(1).run()  # .run() from `statement`

# -- Block 4 --
# `script` details as read from 'overview.sql'
print(script.depth)   #> 7
print(script(1).nm)   #> create table~s1
print(script(-1).nm)  #> drop table~s7

with script.filter(excl_kw=['select', 'drop']) as s:
    print(s.depth)   #> 4
    print(s(1).nm)   #> create table~s1
    print(s(-1).nm)  #> insert into~s4
    s.dtl()
"""
>>>
overview.sql
============
1: Statement('create table~s1')
2: Statement('insert into~s2')
3: Statement('create transient table~s3')
4: Statement('insert into~s4')
"""
# snowmobile-include



"""Demonstrate primary methods for executing raw sql.
../docs/snippets/snowmobile/executing.py
"""
import snowmobile

sn = snowmobile.connect()

# -- sn.query() ---------------------------------------------------------------
df = sn.query("select 1")  #  == pd.read_sql()
type(df)                   #> pandas.core.frame.DataFrame

# -- pd.read_sql() --
import pandas as pd

df2 = pd.read_sql(sql="select 1", con=sn.con)

print(df2.equals(df))  #> True


# -- sn.ex() ------------------------------------------------------------------
cur = sn.ex("select 1")    #  == SnowflakeConnection.cursor().execute()
type(cur)                  #> snowflake.connector.cursor.SnowflakeCursor

# -- SnowflakeConnection.cursor().execute() --
cur2 = sn.con.cursor().execute("select 1")

print(cur.fetchone() == cur2.fetchone())  #> True


# -- sn.exd() -----------------------------------------------------------------
dcur = sn.exd("select 1")  #  == SnowflakeConnection.cursor(DictCursor).execute()
type(dcur)                 #> snowflake.connector.DictCursor

# -- SnowflakeConnection.cursor(DictCursor).execute() --
from snowflake.connector import DictCursor

dcur2 = sn.con.cursor(cursor_class=DictCursor).execute("select 1")

print(dcur.fetchone() == dcur2.fetchone())  #> True

# -- complete example; should run 'as is' --
# snowmobile-include



"""
Demonstrate calling .connect() on existing Snowmobile instances.
..docs/snippets/snowmobile/connector_delayed2.py
"""
import snowmobile

# -- Delayed Connection --
sn_del = snowmobile.connect(delay=True)

print(type(sn_del.con))  #> None
sn_del.connect()
print(type(sn_del.con))  #> snowflake.connector.connection.SnowflakeConnection


# -- Live Connection --
sn_live = snowmobile.connect()

session1 = sn_live.sql.current('session')
sn_live.connect()
session2 = sn_live.sql.current('session')
print(session1 != session2)  #> True

# -- complete example; should run 'as is' --
# snowmobile-include



"""
Instantiate a vanilla Snowmobile and inspect key attributes.
../docs/snippets/snowmobile/inspect_connector.py
"""
import snowmobile

sn = snowmobile.connect()

type(sn)         #> snowmobile.core.connection.Snowmobile

type(sn.cfg)     #> snowmobile.core.configuration.Configuration
str(sn.cfg)      #> snowmobile.Configuration('snowmobile.toml')

type(sn.con)     #> snowflake.connector.connection.SnowflakeConnection
type(sn.cursor)  #> snowflake.connector.cursor.SnowflakeCursor

# -- complete example; should run 'as is' --
# snowmobile-include



"""
Contains parsing utilities used by:

-   :class:`snowmobile.Script`
-   :class:`snowmobile.core.Statement`
-   :class:`snowmobile.core.Markup`
-   :class:`snowmobile.core.Section`

"""
from typing import List, Optional, Tuple, Dict


def rmerge_dicts(d1: Dict, d2: Dict) -> Dict:
    """Recursively merges dictionaries ``d1`` & ``d2``.

    Used for combining contents of **snowmobile.toml** and
    **snowmobile_ext.toml** at the appropriate level within a pair of nested
    dictionaries.

    .. note::
        The order of ``d1`` and ``d2`` matters; if an overlapping key at the
        same level exists in both dictionaries, the value within ``d1``
        will take precedent over the value in ``d2``.

        This function is invoked within the *__init__()* method of
        :class:`~snowmobile.core.Configuration` to merge ``snowmobile.toml``
        and ``snowmobile_ext.toml``, and the prioritization of overlapping
        values described above is the reason that specifications in
        ``snowmobile.toml`` will take precedent over any overlapping
        configurations in ``snowmobile.toml``.

    Args:
        d1 (dict): First dictionary to merge.
        d2 (dict): Second dictionary to merge.

    Returns (dict):
        Dictionary containing recursively merged contents from ``d1`` and
        ``d2``

    """
    merged = d1.copy()
    merged.update(
        {
            key: rmerge_dicts(merged[key], d2[key])
            if (isinstance(merged.get(key), dict) and isinstance(d2[key], dict))
            # else d2[key]
            else (d2[key] if not d1.get(key) else d1[key])
            for key in d2.keys()
        }
    )
    return merged


def up(nm: str):
    """Utility to truncate upper-casing strings as opposed to str.upper()."""
    return nm.upper() if nm else nm


def strip(
    val: str, trailing: bool = True, blanks: bool = True, whitespace: bool = False
) -> str:
    """Utility to strip a variety whitespace from a string."""
    splitter = val.split("\n")
    if trailing:
        splitter = [v.strip() for v in splitter]
    if blanks:
        splitter = [v for v in splitter if v and not v.isspace()]
    if whitespace:
        splitter = [v for v in splitter if not v.isspace()]
    return "\n".join(splitter)


# -- arg_to_list test cases
# input_to_expected = [
#     (
#         """['.*_ignore', "ignore.*",'dummy pattern']""",
#         ['.*_ignore', 'ignore.*', 'dummy pattern']
#     ),
# ]
# for item in input_to_expected:
#     input, expected = item
#     assert arg_to_list(input) == expected


def dict_flatten(
    attrs: dict,
    delim: Optional[str] = None,
    indent_char: Optional[str] = None,
    bullet_char: Optional[str] = None,
) -> List[Tuple[str, str, str]]:
    """Recursively flattens dictionary to its atomic elements.

    Flattens a dictionary to its atomic state and performs parsing operations,
    separating each child key with a `delim` relative to its parent key.

    This flattening enables the parsing of a nested dictionary into a valid
    string of markdown containing correspondingly nested bullets.

    Args:
        attrs (dict):
            Dictionary of attributes; most likely the :attr:`attrs_parsed`
            from :class:`Statement`.
        delim (str):
            Delimiter to use for separating nested keys; defaults to '~';
        indent_char (str):
            Character to use for indents; defaults to a tab ('\t).
        bullet_char (str):
            Character to use for bullets/sub-bullets; defaults to '-'.

    Returns (List[Tuple[str, str, str]]):
        A list of tuples containing:
            1.  A string of the indentation to use; for 1st-level attributes,
                this will just be the `bullet_char`.
            2.  The fully stratified key, including parents; for 1st-level
                attributes this will mirror the original key that was provided.
            3.  The value of the associated key; this will always mirror the
                value that was provided.

    """
    flattened = list()
    delim = delim or "~"
    c = indent_char or "\t"
    bullet_char = bullet_char or "-"

    def recurse(t, parent_key=""):
        if isinstance(t, dict):
            for k, v in t.items():
                sub_key = f"{parent_key}{delim}{k}"
                recurse(v, sub_key if parent_key else k)
        else:
            depth = len(parent_key.split(delim)) - 1
            indent = f"{c * depth}{bullet_char}" if depth else ""
            flattened.append((indent, parent_key, t))

    recurse(attrs)

    return flattened


def p(nm: str) -> Tuple[str, str]:
    """Utility to parse cfg from dot-prefixed object if included."""
    nm = nm or str()
    partitions = [p for p in nm.partition(".") if p]
    if len(partitions) == 3:
        schema, _, name = partitions
    elif nm.strip().startswith("__"):
        schema, name = nm.strip()[2:], str()
    else:
        schema, name = str(), nm.strip()
    return schema, name



"""
Full configuration object model; represents a parsed ``snowmobile.toml`` file.
"""
__all__ = [
    "Base",
    "Credentials",
    "Connection",
    "Script",
    "Pattern",
    "Marker",
	"Markup",
    "QA",
    "Wildcard",
    "Loading",
    "Put",
    "Copy",
    "SQL",
    "Location",
    "Attributes",
]
from .base import Base
from .connection import Connection, Credentials
from .loading import Loading, Put, Copy
from .sql import SQL
from .extensions import Location
from .script import QA, Attributes, Markup, Marker, Pattern, Script, Wildcard



# import link
from . import link

link_name = "bulk loading from a local file system"
user_text = "Bulk Loading from a Local File System"
url = "https://docs.snowflake.com/en/user-guide/data-load-local-file-system.html"

link.xref_links.update({link_name: (user_text, url)})



"""
Instantiate a delayed snowmobile.Snowmobile object.
../docs/snippets/instantiate_connector.py
"""

import snowmobile

sn = snowmobile.Snowmobile(delay=True)

type(sn)                  #> snowmobile.core.connection.Snowmobile
type(sn.con)              #> None

type(sn.cfg)              #> snowmobile.core.configuration.Configuration
print(sn.cfg.location)    #  'path/to/your/snowmobile.toml'

type(sn.cfg.connection)   #> snowmobile.core.cfg.connection.Connection
type(sn.cfg.loading)      #> snowmobile.core.cfg.loading.Loading
type(sn.cfg.script)       #> snowmobile.core.cfg.script.Script
type(sn.cfg.sql)          #> snowmobile.core.cfg.other.SQL
type(sn.cfg.ext_sources)  #> snowmobile.core.cfg.other.Location

# -- complete example; should run 'as is' --



"""Tests for snowmobile.statement.Diff."""
# import pytest
#
#
# # noinspection SqlResolve
# @pytest.fixture()
# def sample_diff_object(sn):
#     """An example statement object for testing."""
#     from snowmobile.core.statement import Diff
#
#     # creating an example table for fixture
#     sn.ex('drop table if exists an_example_table')
#     sn.ex('create temp table an_example_table as select 1 as sample_col')
#
#     # sql for statement object
#     sql = """
# select
#     *
# from an_example_table a;
# """
#
#     return Statement(
#         sn=sn,
#         statement=sn.cfg.script.ensure_sqlparse(sql=sql.strip('\n')),
#         index=1
#     )



# -*- coding: utf-8 -*-
"""
*   External links per the first link below
*   This is supposed to be included in the 'sphinx.ext.extlinks', but it
    doesn't work out of the box and super hypocritically, the author of the
    documentation (first link below) outlines using the 'extlinks' extension
    and references the repo for the project as an example to follow (second
    link below), but within that repo, he's not using the builtin sphinx
    extension referenced in the documentation and instead is embedding his own
    extension which is the below set of codes.

* Sphinx docs on extlinks
    *   https://sublime-and-sphinx-guide.readthedocs.io/en/latest/references.html
* conf.py of the sublime-and-sphinx project
    *   https://github.com/MarkHoeber/sublime_sphinx_guide/blob/master/source/conf.py
*
"""

from docutils import nodes

from sphinx.util import caption_ref_re


def xref(typ, rawtext, text, lineno, inliner, options={}, content=[]):

    title = target = text
    titleistarget = True
    # look if explicit title and target are given with `foo <bar>` syntax
    brace = text.find("<")
    if brace != -1:
        titleistarget = False
        m = caption_ref_re.match(text)
        if m:
            target = m.group(2)
            title = m.group(1)
        else:
            # fallback: everything after '<' is the target
            target = text[brace + 1 :]
            title = text[:brace]

    link = xref.links[target]

    if brace != -1:
        pnode = nodes.reference(target, title, refuri=link[1])
    else:
        pnode = nodes.reference(target, link[0], refuri=link[1])

    return [pnode], []


def get_refs(app):

    xref.links = app.config.xref_links


def setup(app):

    app.add_config_value("xref_links", {}, True)
    app.add_role("xref", xref)
    app.connect("builder-inited", get_refs)



"""
Module contains the object model for **snowmobile.toml**.
"""
from __future__ import annotations

from typing import Any, Iterable, Set, Type

from .errors import *
from . import Generic


class ExceptionHandler(Generic):
    """Context management for snowmobile objects."""

    def __init__(
        self,
        within: Optional[Any] = None,
        ctx_id: Optional[int] = None,
        in_context: bool = False,
        children: Dict[int, Any] = None,
        is_active_parent: bool = False,
        to_mirror: Optional[List[Any]] = None,
    ):
        super().__init__()
        self._ctx_id: Optional[int] = ctx_id

        self.within: Type = type(within) if within else None
        self.by_ctx: Dict[int, Dict[int, snowmobile_errors]] = dict()
        self.in_context: bool = in_context
        self.outcome: Optional[int] = None
        self.children = children
        self.is_active_parent: bool = is_active_parent
        self.to_mirror: Optional[List[Any]] = to_mirror

    @property
    def ctx_id(self):
        """Current context id."""
        if not self._ctx_id:
            self.set(ctx_id=time.time_ns())
        return self._ctx_id

    def set(
        self,
        ctx_id: Optional[int] = None,
        in_context: bool = False,
        outcome: Optional[int] = None,
    ):
        """**Set** attributes on self."""
        if ctx_id:
            if ctx_id in self.by_ctx and self:
                nm = self.within.__name__
                raise InternalError(
                    nm=f"ExceptionHandler.set() from within {nm}",
                    msg=f"an existing `_ctx_id`, {ctx_id}, was provided to `set(_ctx_id)`",
                )
            self._ctx_id = ctx_id if ctx_id != -1 else time.time_ns()
            self.by_ctx[self._ctx_id] = {}
        if in_context:
            self.in_context = in_context
        if outcome:
            self.outcome = outcome

        if self.is_active_parent and "set" in self.to_mirror:
            for child in self.children.values():
                child.e.set(ctx_id=ctx_id, in_context=in_context, outcome=outcome)

        return self

    def set_from(self, other: ExceptionHandler) -> ExceptionHandler:
        """Updates attributes from another ExceptionHandler object."""
        for k in ["_ctx_id", "in_context", "outcome"]:
            vars(self)[k] = vars(other)[k]
        return self

    def reset(
        self, ctx_id: bool = False, in_context: bool = False, outcome: bool = False
    ):
        """**Reset** attributes on self."""
        if ctx_id:
            self._ctx_id = None
        if in_context:
            self.in_context = False
        if outcome:
            self.outcome = None

        if self.is_active_parent and "reset" in self.to_mirror:
            for child in self.children.values():
                child.e.reset(ctx_id=ctx_id, in_context=in_context, outcome=outcome)

        return self

    @property
    def current(self):
        """All exceptions in the current context."""
        if not self._ctx_id:
            raise InternalError(
                nm="ExceptionalHandler.current",
                msg=f"""
A call was made to `.current` while the current value of '_ctx_id` is None.
""".strip(
                    "\n"
                ),
            )
        return self.by_ctx[self.ctx_id] if self.ctx_id in self.by_ctx else {}

    def collect(self, e: Any[snowmobile_errors]):
        """Stores an exception."""
        current = self.current
        current[int(time.time_ns())] = e
        self.by_ctx[self.ctx_id] = current
        return self

    def _first_last(self, idx: int):
        """Last exception encountered."""
        by_tmstmp = self.by_tmstmp
        return by_tmstmp[list(by_tmstmp)[idx]] if by_tmstmp else {}

    @property
    def first(self) -> Error:
        """First exception encountered."""
        return self._first_last(-1)

    @property
    def last(self) -> Error:
        """Last exception encountered."""
        return self._first_last(0)

    @staticmethod
    def _query_types(
        to_search: Dict[int, snowmobile_errors],
        of_type: Optional[snowmobile_errors, List[snowmobile_errors]] = None,
    ) -> Set[int]:
        """Search through exceptions by type."""
        of_type = of_type if isinstance(of_type, Iterable) else [of_type]
        return {
            e for e in to_search if any(isinstance(to_search[e], t) for t in of_type)
        }

    @staticmethod
    def _query_mode(
        to_search: Dict[int, snowmobile_errors], to_raise: bool
    ) -> Set[int]:
        """Search through exceptions by mode (to_raise=True/False)."""

        def _raise(e: Any):
            """Generalized check of an assertion's intention to be raised."""
            return getattr(e, "to_raise") if hasattr(e, "to_raise") else True

        return {
            _id
            for _id, e in to_search.items()
            if (_raise(e) if to_raise else not _raise(e))
        }

    def _query(
        self,
        of_type: Optional[snowmobile_errors, List[snowmobile_errors]] = None,
        to_raise: Optional[bool] = None,
        with_ids: Optional[int, List[int], Set[int]] = None,
        from_ctx: Optional[int] = None,
        all_time: bool = False,
    ) -> Dict[int, snowmobile_errors]:
        """Search through exceptions encountered by criterion."""
        to_consider: Dict[int, snowmobile_errors] = (
            self.by_ctx[from_ctx or self.ctx_id] if not all_time else self.by_tmstmp
        )
        seen = set(to_consider)

        if of_type:
            seen = seen.intersection(
                self._query_types(to_search=to_consider, of_type=of_type)
            )

        if isinstance(to_raise, bool):
            seen = seen.intersection(
                self._query_mode(to_search=to_consider, to_raise=to_raise)
            )

        if with_ids:
            seen = seen.intersection(
                set(with_ids if isinstance(with_ids, Iterable) else [with_ids])
            )

        return {i: to_consider[i] for i in sorted(seen, reverse=True)} if seen else {}

    def seen(
        self,
        from_ctx: Optional[int] = None,
        of_type: Optional[Any[snowmobile_errors], List[snowmobile_errors]] = None,
        to_raise: Optional[bool] = None,
        with_ids: Optional[int, List[int], Set[int]] = None,
        all_time: bool = False,
    ) -> bool:
        """Boolean indicator of if an exception has been seen."""
        return bool(
            self._query(
                of_type=of_type,
                to_raise=to_raise,
                with_ids=with_ids,
                from_ctx=from_ctx,
                all_time=all_time,
            )
        )

    def get(
        self,
        from_ctx: Optional[int] = None,
        of_type: Optional[Any[snowmobile_errors], List[snowmobile_errors]] = None,
        to_raise: Optional[bool] = None,
        with_ids: Optional[int, List[int], Set[int]] = None,
        all_time: bool = False,
        last: bool = False,
        first: bool = False,
        _raise: bool = False,
    ):
        """Boolean indicator of if an exception has been seen."""
        encountered = self._query(
            of_type=of_type,
            to_raise=to_raise,
            with_ids=with_ids,
            from_ctx=from_ctx,
            all_time=all_time,
        )
        if last and not encountered:
            raise InternalError(
                nm="ExceptionHandler.get()",
                msg=f"""
a call was made to `.get()` that returned no exceptions;
exceptions in current context are:\n\t{list(self.current.values())}
""".strip(
                    "\n"
                ),
            )

        if (last or first) and encountered:
            return encountered[list(encountered)[0 if last else -1]]
        return encountered

    @property
    def by_tmstmp(self):
        """All exceptions by timestamp, ordered by most to least recent."""
        unordered = {k2: v2 for k, v in self.by_ctx.items() for k2, v2 in v.items()}
        return {k: unordered[k] for k in sorted(unordered, reverse=True)}

    def __len__(self):
        return len(self.by_tmstmp)

    def __bool__(self):
        """False=no exceptions to be raised in current context."""
        return self.seen(to_raise=True)

    def __repr__(self):
        return (
            f"ExceptionHandler(within={self.within}, contexts={len(self.by_ctx)}, "
            f"cnt={len(self)}, to_raise: {bool(self)}"
        )



"""
This file populates snippets.md based on a source directory and the parameters
defined within this file directly below.

It is not under test and was written to serve a single purpose of generating
this reference page.
"""
from typing import List, Dict, Tuple
from pathlib import Path

# --/ Hard Codes /-------------------------------------------------------------

GENERIC_DIR_NAME = "General"  # header for snippets in base directory
FLAG = 'snowmobile-include'  # string that indicates the include the file

OFFSET_TOP = 0  # offset from top of files to start
OFFSET_BOTTOM = 2  # offset from bottom of files to start

TITLE_CASE_DIR_NAMES = True  # whether to title case directory names in headers

MAP_LANGUAGE = {  # maps file extensions to :language: directive
    '.bashrc': 'bash',
    '.bat': 'shell',
    '.html': 'html',
    '.ini': 'ini',
    '.json': 'json',
    '.py': 'python',
    '.rst': 'rst',
    '.sh': 'sh',
    '.sql': 'sql',
    '.toml': 'toml',
    '.R': 'R',
    '.xml': 'xml',
    '.yml': 'yaml'
}

# HERE = Path.cwd() / 'docs' / 'ext' / 'snippets.py'
HERE = Path(__file__).absolute()  # current file location
DOC_ROOT = HERE.parent.parent     # documentation root directory
SNIPPETS_DIR_NAME = 'snippets'    # directory containing code snippets

TARGET_FILE_NAME = "snippets.md"  # file name to save raw text in

DIV_WRAP = True  # wrap {literalinclude} directive in a div
DIV_CLASSES = ['sn-indent-h-cell']  # classes for div


# INCL_HR = True
INCL_HRS = ['###']
INCL_LINE_NUMBERS = True

ALT_HEADERS = {
    'h2': {'sql': 'SQL'}
}

GIT_ROOT = r'https://github.com/GEM7318/Snowmobile/tree/0.2.1/docs'
DIVS_DOWNLOAD = ['sn-download-snippet']
DIVS_GIT_LINK = ['sn-snippet-git-link']
DIVS_PARENT_CONTAINER = ['sn-link-group', 'sn-snippet-only']

INCL_DOWNLOAD = True
INCL_GIT_LINK = True
INCL_PARENT_CONTAINER = True  # must be True if either _DOWNLOAD or _GIT_LINK is True


# --/ Derived /----------------------------------------------------------------
offsets = (OFFSET_TOP, OFFSET_BOTTOM)
root = DOC_ROOT / 'snippets'
target_file_path = DOC_ROOT / TARGET_FILE_NAME


def line_nos(txt: str, offsets: Tuple[int, int]) -> Tuple[int, int]:
    """Determines lines to include based on the raw text and offsets.

    Args:
        txt:
            Raw file contents.
        offsets:
            Tuple containing (top-offset, bottom-offset)

    Returns:
        Tuple containing (start-line, end-line)

    """
    _start, _end = offsets
    depth = len(txt.split('\n'))
    start_line = 0 + _start if _start else 1
    end_line = depth - _end
    return start_line, end_line


def download(file_path: Path) -> str:
    """Generates 'download' container"""
    p_rel = list(file_path.relative_to(root).parts)
    p_rel.insert(0, root.name)
    p_download_v0 = '/'.join(p_rel)
    p_download = f"./{p_download_v0}"
    _div = "{div}"
    _download = "{download}"
    return f"""
```{_div} {', '.join(DIVS_DOWNLOAD)} 
{_download}`Download<{p_download}>`
```
""".strip('\n')


def git(file_path: Path) -> str:
    """Generates 'git' container"""
    p_rel = list(file_path.relative_to(root).parts)
    p_rel.insert(0, root.name)
    url_git_v0 = '/'.join(p_rel)
    url_git = f"{GIT_ROOT}/{url_git_v0}"
    _div = "{div}"
    return f"""
```{_div} {', '.join(DIVS_GIT_LINK)} 
<a class="reference external" href="{url_git}"><span>text</span></a>
```
""".strip('\n')


def link_container(file_path: Path) -> str:
    """Generates link container."""
    div_download = ''
    div_git_url = ''
    _div = "{div}"
    if INCL_DOWNLOAD:
        div_download = download(file_path)
    if INCL_GIT_LINK:
        div_git_url = git(file_path)
    return f"""\n
````{_div} {', '.join(DIVS_PARENT_CONTAINER)}
{div_download}
{div_git_url}
````"""


def header(file_path: Path, root: Path):
    """Generates literal header text from a file path within a root path."""
    offset = list(file_path.relative_to(root).parts)[:-1]
    h_prefix = (2 + len(offset)) * '#'
    h_final = f"{h_prefix} *{file_path.name}*"
    return h_final if h_prefix not in INCL_HRS else f"{h_final}\n---"


def rst_literal_incl(
    file_path: Path,
    root: Path,
    offsets: Tuple[int, int],
    flag: str,
    suffix_map: Dict[str, str],
    first_of_dirs: Dict[str, Path],
    title_dir_names: bool = TITLE_CASE_DIR_NAMES
) -> str:
    """DOCSTRING"""
    with open(file_path, 'r') as r:
        raw = r.read()

    parent_header = str()
    parent = file_path.parent.name

    if flag not in raw:
        return str()

    if parent in first_of_dirs:
        _ = first_of_dirs.pop(parent)
        alt_h2 = ALT_HEADERS.get('h2')
        parent = alt_h2.get(
            parent.lower(),
            parent.title() if title_dir_names else parent
        )
        parent_header = f"<br>\n\n## {parent}"
    child_header = header(file_path, root)
    header_final = (
        f"{parent_header}\n\n{child_header}" if parent_header
        else child_header
    )

    path_sub = '/'.join(file_path.relative_to(root).parts)
    path_final = f"/{root.name}/{path_sub}"

    start_line, end_line = line_nos(raw, offsets)
    language = suffix_map[file_path.suffix]
    incl = "{literalinclude}"
    lines = '' if not INCL_LINE_NUMBERS else "\n:lineno-start: 1"

    literal_include = (
        f"""
```{incl} {path_final}
:language: {language}
:lines: {start_line}-{end_line}{lines}
```
"""
    )

    if DIV_WRAP:
        _div = "{div}"
        _classes = ', '.join(DIV_CLASSES)
        div = f"{_div} {_classes}"
        literal_include = f"````{div}{literal_include}````"

    _links = ''
    if INCL_PARENT_CONTAINER:
        _links = link_container(file_path)

    return (
        f"""
{header_final}{_links}
{literal_include}
<br>
"""
    )


def get_first_of_dir(files_sorted: Dict[str, Path]):
    """DOCSTRING"""
    first_of_dir = {}
    for f in files_sorted.values():
        if f.parent.name not in first_of_dir.values():
            first_of_dir[f.parent.name] = f
    return first_of_dir


def snippets_total(
    files: List[Path],
    root: Path,
    offsets: Tuple[int, int],
    flag: str,
    suffix_map: Dict[str, str],
    first_of_dirs: Dict[str, Path],
):
    """docstring"""
    print("----/ Starting Snippets /----")
    total_rst = []
    for f in files:
        rst = rst_literal_incl(
            file_path=f,
            root=root,
            offsets=offsets,
            flag=flag,
            suffix_map=suffix_map,
            first_of_dirs=first_of_dirs,
        )
        if rst:
            total_rst.append(rst)
            print(f"\tsnippet: {f.name}")
    return total_rst


def write():
    """docstring"""
    files = {
        f.as_posix(): f
        for f in root.rglob('**/*.*')
        if f.is_file() and f.suffix in MAP_LANGUAGE
    }

    files_sorted = {files[k].name: files[k] for k in sorted(files)}
    first_of_dirs = get_first_of_dir(files_sorted)

    total = snippets_total(
        files=list(files_sorted.values()),
        root=root,
        offsets=offsets,
        flag=FLAG,
        suffix_map=MAP_LANGUAGE,
        first_of_dirs=first_of_dirs
    )

    with open(target_file_path, 'r') as r:
        current = r.read()

    page_heading, _ = current.split("+++")
    to_write = f"{page_heading}+++{''.join(total)}"

    # to_write = f"{PAGE_HEADER}{''.join(total)}"
    # if ADD_STYLE:
    #     to_write = f"{to_write}\n\n{PAGE_STYLE}"

    with open(target_file_path, 'w') as f:
        f.write(to_write)



"""Tests for snowmobile.Statement."""
import pytest


# noinspection SqlResolve
@pytest.fixture()
def sample_statement_object(sn):
    """An example statement object for testing."""
    from snowmobile import Statement

    # creating an example table
    sn.ex("drop table if exists an_example_table")
    sn.ex("create temp table an_example_table as select 1 as sample_col")

    # sql for statement object
    sql = """
select
    *
from an_example_table a;
"""
    return Statement(
        sn=sn, statement=sn.cfg.script.ensure_sqlparse(sql=sql.strip("\n")), index=1
    )


@pytest.mark.statement
def test_statement_number_of_lines(sample_statement_object):
    """Tests the statement.lines property."""
    assert sample_statement_object.lines == 3


@pytest.mark.statement
def test_multiline_attributes_without_a_name_causes_error(sample_statement_object):
    """Verifies that a multiline tag without a '__name:' argument will raise an error."""
    from snowmobile.core.errors import InvalidTagsError

    multiline_tag_without_a_name_argument = """
__description: This is an example of a multiline tag that will cause an error.
"""

    # given
    sample_statement_object.attrs_raw = multiline_tag_without_a_name_argument
    sample_statement_object.is_multiline = True
    sample_statement_object.is_tagged = True

    # then
    with pytest.raises(InvalidTagsError):
        sample_statement_object.parse()


@pytest.mark.statement
def test_calling_render_on_a_valid_statement_does_not_cause_an_error(
    sample_statement_object
):
    """Tests statement.render() method does not cause an error."""
    sample_statement_object.run(render=True)


# noinspection SqlResolve
@pytest.mark.statement
def test_calling_run_with_results_on_invalid_sql_raises_database_error(
    sample_statement_object
):
    """Tests `s.run(as_df=True)` raises an error when `s.sql` is not a valid statement."""
    from pandas.io.sql import DatabaseError

    # given
    an_invalid_sql_statement = "select * from *"
    sample_statement_object.set_sql(sql=an_invalid_sql_statement)

    # then
    with pytest.raises(DatabaseError):
        sample_statement_object.run(as_df=True)


@pytest.mark.statement
def test_statement_dunder_getitem(sample_statement_object):
    """Verifies `s.__getitem__()`."""
    assert (
        # sample_statement_object.__getitem__(item="start_time") == sample_statement_object.sql
        sample_statement_object.__getitem__(item="start_time") == sample_statement_object.start_time
    )


@pytest.mark.statement
def test_statement_dunder_str(sample_statement_object):
    """Verifies `s.__str__()`."""
    assert str(sample_statement_object) == "Statement('select data~s1')"
    # assert str(sample_statement_object)


@pytest.mark.statement
def test_statement_dunder_setitem(sample_statement_object):
    """Verifies `s.__str__()`."""
    sample_statement_object.__setitem__("start_time", 1)
    assert sample_statement_object.start_time == 1
